"""
This module define a class and tools to efficiently solve large
linear problems in python.

The difficulty to efficiently go through large dataset in pure python is
solved through the use of scipy.sparse to make the actual
computation. For the same price we get the possibility of solving very
large sparse system :)

The code now handle outliers rejection with rank-one update of the
cholesky decomposition. It was also significantly refactored with a new
and incompatible API.

TODO
----
- Use studentized residuals in the outliers detection
- Wrap spqr
- Implement linear equality constrainsts
- Test resolution with orthogonal factorization of A
- Damping ?
"""

import copy
import logging

import numpy as np
import selinv
from scipy import sparse
from sksparse.cholmod import cholesky, cholesky_AAt

from .fitparameters import FitParameters, structarray
from .indextools import make_index
from .robuststat import binned_curve, polyfit

logger = logging.getLogger(__name__)

# pylint: disable=invalid-name,attribute-defined-outside-init,protected-access


class SparseSystem(object):
    def __init__(self, A, b, **keys):
        self.b = b
        self.keys = keys
        self.factor = cholesky_AAt(A.T, **keys)

    def update(self, A, b, subtract=True):
        self.factor.update_inplace(A.T, subtract=subtract)
        if subtract:
            self.b -= b
        else:
            self.b += b

    def refact(self, A, b):
        self.b = b
        self.factor.cholesky_AAt_inplace(A.T)

    def solve(self):
        x = self.factor(self.b)
        try:
            return x.squeeze()
        except AttributeError:
            return x

    def fullinv(self):
        n = self.factor.D().shape[0]
        D = sparse.eye(n, n).tocsc()
        return self.factor(D)


class SparseLocalGlobalSystem(object):
    """When a very big block of the Hessian is known to be diagonal, we
    can take explicitly advantage of that structure to speed up things
    rather than resorting to the generic sparse factorisation.

    """

    def __init__(self, A, b, n):
        self.b = b
        self.n = n
        A1 = A[:, : self.n]
        A2 = A[:, self.n :]
        self.D = (A1.T * A1).diagonal()
        self.B = A1.T * A2
        self.M = A2.T * A2
        # self.M = A.T * A
        self._fill()

    def _fill(self):
        self.b1 = self.b[: self.n]
        self.b2 = self.b[self.n :]
        # self.Dinv = 1./(A1.T*A1).diagonal()
        self.Dinv = 1.0 / self.D
        # self.B = A1.T * A2
        self.Bp = self.B.copy()
        self.Bp.data *= self.Dinv[self.Bp.indices]  # self.Bp = self.Dinv * B
        self.S = self.M - self.B.T * self.Bp
        # self.Mfact = cholesky(A2.T * A2 - self.B.T * self.Bp)
        self.Mfact = cholesky(self.S)

    def solve(self):
        x1 = self.Dinv * self.b1
        x2 = self.Mfact(self.b2)
        # x3 = self.Bp * x2
        x4 = self.Mfact(self.B.T * x1)
        return np.hstack([x1 + self.Bp * (x4 - x2), -x4 + x2])

    def update(self, W, b, subtract=True):
        # For now we redo the factorization, but a true update might be envisionned
        W1, W2 = W[:, : self.n], W[:, self.n :]
        if subtract:
            D = -(W1.T * W1)
            DeltaB = -W1.T * W2
            self.S -= W2.T * W2
            self.b -= b
        else:
            D = W1.T * W1
            DeltaB = W1.T * W2
            self.S += W2.T * W2
            self.b += b
        self.b1 = self.b[: self.n]
        self.b2 = self.b[self.n :]
        nz = D.indices
        am1 = self.Dinv[nz]
        bovera = am1 * D.data
        Dw = -am1 / (1 + bovera) * bovera
        # A verifier D'' = D + Dw
        self.Dinv[nz] += Dw
        # assert (np.abs(self.Dinv * (self.D + D.diagonal())-1) < 1e-15).all()
        DeltaBp = DeltaB.copy()
        DeltaBp.data *= self.Dinv[DeltaBp.indices]
        D.data = Dw
        DeltaBp += D * self.B
        self.S -= self.B.T * DeltaBp
        self.Bp += DeltaBp
        self.B += DeltaB
        self.S -= DeltaB.T * self.Bp
        self.Mfact.cholesky_inplace(self.S)

    def refact(self, A, b):
        """Same as above, restarting from scratch"""
        self.M = A.T * A
        self.b = b
        self._fill()

    def fullinv(self):
        n = self.b.shape[0]
        D = sparse.eye(n, n).tocsc()
        return self.factor(D)

    def diaginv(self):
        # Reusing the cholmod factorization is hard. In contrast
        # reusing the reordering is easy and is important for
        # performances because cholmod metis ordering is typically
        # better than the default MMD ordering included in selinv.
        Adiag = self.D
        return selinv.selinv(self.S, self.Mfact.P())


class RobustLinearSolver(object):
    r"""Direct solver for sparse linear models with iterative outlier rejection

    Basics:
    -------

    1) This class provides the function "solve" to compute the
    Ordinary Least Square (OLS) estimator \hat{X} for the linear
    model:

    $$y = Ax + n$$,

    by the closed-form relation:

    $$\hat{x} = (A^T A)^{-1} A^T y.$$

    Both the computation of A^T A and A^T y and the resolution of the
    linear system are made using the sparse matrix package from scipy.

    2) Optionnaly one can provide a vector of weights w in which case
    the Weighted Least Square problem is first transformed to a OLS
    problem by the following modifications:

    $$y \rightarrow \diag(w) y $$

    and

    $$A \rightarrow \diag(w) A$$

    2bis) an alternative is to provide the class with a non-diagonal
    *sparse* weight matrix. In this case, the Cholesky decomposition
    of W is determined, W = LL' and the WLS is transformed as follows:

    $$y \rightarrow L' & y$$

    and

    $$A \rightarrow L' A$$

    3) The function "robust_solution" performs outlier rejections,
    solving for the screening problem iteratively. The flagging of
    outliers is controlled by the flag_outliers function.

    The default implementation use a careful algorithm well adapted to
    problems with a lot of local parameters, but which may converge too
    slowly for other problems. The class can be derived easily to
    reimplement the flag_outliers function.

    4) get_cov compute the Fisher matrix of a subset of best-fit
    parameters.

    5) get_diag_cov compute the diagonal of the Fisher matrix of
    best-fit parameters in an efficient (sparse) way.

    6) diagnostic_plot is an utilitary function that displays the
    distribution of residuals and their dispersion as a function of
    given variables.

    7) weight_model is an utilitary function that fit a polynomial P
    of an explanatory variable v to the empirical rms of residuals in
    bins of v.

    8) reweight fit and apply a weight model to the problem
    """

    def __init__(self, model, y, weights=None, verbose=1, nlocal=0, **keys):
        self.verbose = verbose

        self.A = model.get_free_mat()
        self.y = model.get_free_y(y)
        self.model = model

        # Apply weights if needed
        if weights is not None:
            if weights.ndim == 1:
                assert len(weights) == len(y), (
                    "The lenght of the provided weight vector differs"
                    " from the number of measurements"
                )
                assert np.isfinite(weights).all(), (
                    "Some weights have non finite values,"
                )
                self.A = self.A.tocoo()
                self.A.data *= weights[self.A.row]
                # self.y is already initialized to a copy of y
                # self.y = y * weights
                self.y *= weights
                # else not needed: self.y is already initialized to a copy of y
                #        else:
                #            self.y = y
            elif issubclass(type(weights), sparse.spmatrix):
                shape = weights.shape
                assert shape[0] == shape[1] == len(y), (
                    "The length of thr provided weight vector differs"
                    " from the number of measurements"
                )
                fact = cholesky(weights)
                Lt = fact.L().T
                self.A = Lt * self.A
                self.y = Lt * self.y

        # Cholesky decomposition
        self.A = self.A.tocsr()
        if nlocal:
            self.system = SparseLocalGlobalSystem(self.A, self.A.T * self.y, nlocal)
        else:
            self.system = SparseSystem(self.A, self.A.T * self.y, **keys)
        self.bads = None

    def suppress_one_at_a_time(self, r, local_param):
        """For each parameter, suppress only the worst outlier

        This is hardly maintainable. Did not find an efficient yet readable way to write this.
        """
        suppress_one = np.zeros_like(self.suppress[self.suppress])
        # Get the indices of entries corresponding to outliers
        rows, cols = self.A[self.suppress.nonzero()[0], :].nonzero()
        # Extract the residuals corresponding to outliers
        rsuppress = r[self.suppress]
        # For each parameter (column) store the list of corresponding entries in rsuppress
        index = make_index(cols)
        # For each parameter in the set of local_param, go through the
        # list of corresponding residuals, find the worst and put the
        # corresponding row in a set
        indexes = self.model.params[local_param].indexof()
        indexes = indexes[indexes != -1]
        start = indexes.min()
        stop = indexes.max()
        worst = list(
            set(
                [
                    rows[ind][np.argmax(rsuppress[rows[ind]])]
                    for ind in index
                    if (cols[ind[0]] >= start) and (cols[ind[0]] <= stop)
                ]
            )
        )
        # raise ValueError()
        # Not sure the sorting is necessary
        worst.sort()
        # Now we indeed suppress the rows corresponding to the worst
        # outlier in each column
        suppress_one[worst] = True
        self.suppress[self.suppress] = suppress_one
        # print np.bincount(self.A[self.suppress.nonzero()[0], :].nonzero()[1]).max()
        # print np.where(np.bincount(self.A[~self.suppress, :].nonzero()[1]) == 0)[0]

    def robust_solution(self, **keys):
        """Compute a robust solution to the linear system, iteratively discarding
        outliers.

        Parameters:
        -----------
        keys: keyword arguments are passed directly to self.flag_outliers.
        """
        x = self.system.solve()
        self.flag_outliers(x, **keys)
        niter = 1
        while self.suppress is not None or self.restore is not None:
            if self.suppress is not None:
                C = self.A[self.suppress.nonzero()[0], :]
                self.system.update(C, C.T * self.y[self.suppress])
            if self.restore is not None:
                C = self.A[self.restore.nonzero()[0], :]
                self.system.update(C, C.T * self.y[self.restore], subtract=False)
            x = self.system.solve()
            self.flag_outliers(x, **keys)
            niter += 1
        if self.verbose:
            Nout = self.bads.sum()
            N = float(len(self.bads))
            logger.debug(
                (
                    "Converged in %d iter, cutting %d/%d (%.1g%%) of the measurements"
                    % (niter, Nout, N, Nout / N * 100)
                )
            )
        return x

    def flag_outliers(self, x, nsig=5, local_param=None, ensure_finite=False):
        """Given a solution x, flag measurements whose residuals to the fit

        deviates by more than nsig sigma from zero. The standard
        deviation of the residuals is evaluated as the square root of
        the chi2 divided by the number of degrees of freedom.

        If local_param is provided, make sure that for each local
        parameter at most one measurement affecting this parameter is
        suppressed.

        This methods uses the provided x and self.bads as starting
        points. It sets self.suppress, self.restore and self.bads.

        """
        # sometimes, x is a numpy scalar - ensure that we have an array
        x = np.atleast_1d(x)
        r = (self.A * x).squeeze()
        r = np.atleast_1d(r)
        r -= self.y
        r *= r
        if self.bads is not None:
            chi2 = np.sum(r[~self.bads])
            Nmeas = len(r) - np.sum(self.bads)
        else:
            chi2 = np.sum(r)
            Nmeas = len(r)
        if ensure_finite:
            assert np.isfinite(chi2)
        dof = max(Nmeas - self.A.shape[1], 1)

        if self.verbose:
            med_chi2 = np.nanmedian(r)
            logger.debug(
                (
                    "Chi2: %f, D.o.F.: %d, Chi2/DoF: %f, median Chi2: %f"
                    % (chi2, dof, chi2 / dof, med_chi2)
                )
            )

        bads = r > (nsig**2 * chi2 / dof)

        if self.bads is not None:
            self.suppress = bads & ~self.bads
            self.restore = self.bads & ~bads
        else:
            self.suppress = bads
            self.restore = [False]
        if local_param is not None and self.suppress.any():
            self.suppress_one_at_a_time(r, local_param)

        Nsup = np.sum(self.suppress)
        Nres = np.sum(self.restore)
        if Nsup == 0:
            if self.bads is None:
                self.bads = self.suppress
            self.suppress = None
        else:
            if self.bads is not None:
                self.bads[self.suppress] = True
            else:
                self.bads = self.suppress
        if Nres == 0:
            self.restore = None
        else:
            self.bads[self.restore] = False
        if self.verbose:
            logger.debug("Suppress %d new outliers" % Nsup)
            logger.debug("Restore %d measurements" % Nres)
        self.r = r
        # self.bads = bads
        self.chi2 = chi2

    def get_res(self, y, x=None):
        """Compute the residual"""
        return y - self.model(x)

    def get_wres(self, x=None):
        """Compute weighted residual"""
        if x is None:
            x = self.x
        x = np.atleast_1d(x)
        r = (self.A * x).squeeze()
        r = np.atleast_1d(r)
        r -= self.y
        return r

    def get_cov(self):
        """Compute the full covariance matrix

        Computation of the full inverse of the FIM which can be pretty
        expansive for large problems. See get_diag_cov and get_block_cov
        for a workable alternative in large sparse problems.
        """
        return self.system.fullinv()

    def get_block_cov(self, block):
        """Compute a block of the covariance matrix

        Computation of a block (typically small) of the inverse of the FIM.
        """
        if self.bads is not None:
            M = self.A[(~self.bads).nonzero()[0], :]
        else:
            M = self.A
        nonblock = np.ones(self.A.shape[1], dtype=bool)
        nonblock[block] = False
        nonblock = np.where(nonblock)[0]
        A1 = M[:, nonblock]
        A2 = M[:, block]
        A22 = A2.T * A2
        A12 = A1.T * A2
        bigsys = SparseSystem(A1, A12)
        sol = A12.T * bigsys.solve()
        B = A22 - sol.todense()
        return np.linalg.inv(B)

    # TODO see https://gitlab.in2p3.fr/lemaitre/saltworks/-/issues/1
    # def get_diag_cov(self):
    #     """ Compute the diagonal of the covariance matrix.

    #     This uses the selinv algorithm from Lin Lin et al. 2011 and can
    #     be much faster than computing the full covariance matrix if the
    #     FIM is sparse.
    #     """
    #     if self.bads is not None:
    #         M = self.A[(~self.bads).nonzero()[0], :]
    #     else:
    #         M = self.A
    #     # Reusing the cholmod factorization is hard. In contrast
    #     # reusing the reordering is easy and is important for
    #     # performances because cholmod metis ordering is typically
    #     # better than the default MMD ordering included in selinv.
    #     return selinv.selinv(M.T * M, self.system.factor.P())

    def diagnostic_plot(
        self,
        disperse=[],
        deg=2,
        pbins=None,
        nbins=10,
        data=False,
        keep_outliers=False,
        win=1,
    ):
        """Plot the distribution of residuals.

        Optionnally, plot the dispersion of residuals against the
        provided variables.

        Parameters:
        -----------
        disperse: list
                  the list of arrays to be used as abscisse for the
                  dispersion plots.
        """
        import matplotlib.pyplot as plt
        from scipy.stats import chi2

        from saltworks.plottools import binplot

        # self.r always exists and has the length of the data
        # while self.bads can be None
        if keep_outliers or self.bads is None:
            Nmeas = len(self.r)
        else:
            Nmeas = float(np.sum(~self.bads))
        npars = self.A.shape[1]
        if not disperse:
            plt.figure(win)
            plt.clf()
            if keep_outliers:
                N, bins, h = plt.hist(
                    self.r,
                    bins=100,
                    normed=True,
                    log=True,
                    color="k",
                    histtype="step",
                    label="residuals",
                )
            else:
                N, bins, h = plt.hist(
                    self.r[~self.bads],
                    bins=100,
                    normed=True,
                    log=True,
                    color="k",
                    histtype="step",
                    label="residuals",
                )
            x = 0.5 * (bins[1:] + bins[:-1])
            plt.plot(x, chi2.pdf(x, 1 - npars / Nmeas), "r-", label="chi2 pdf")
            plt.legend()

        for x in disperse:
            #            if len(disperse) > 1:
            #                plt.figure()
            # even if there is only one "disperse" I find it better to have it on its own plot
            plt.figure(win)
            plt.clf()
            if keep_outliers:
                binplot(
                    x,
                    self.r,
                    data=data,
                    marker="s",
                    nbins=nbins,
                    bins=pbins,
                    capsize=0,
                    color="k",
                    label="binned residuals",
                )
            else:
                binplot(
                    x[~self.bads],
                    self.r[~self.bads],
                    data=data,
                    marker="s",
                    nbins=nbins,
                    bins=pbins,
                    capsize=0,
                    color="k",
                    label="binned residuals",
                )
            plt.axhline(
                1 - npars / Nmeas, color="k", ls="-.", label="Expected residual value"
            )
            if pbins is not None:
                w_model = self.weight_model(x, deg=deg, bins=pbins)
                z = np.linspace(*(plt.axis()[:2]))
                plt.plot(
                    z, np.polyval(w_model, z), color="r", label="deg %d model" % (deg)
                )
            plt.legend()

    def weight_model(self, variable, deg=2, bins=10):
        """Todo: Turn that into something proper"""

        # self.bads can be None:
        if self.bads is None:
            x, y, ey = binned_curve(variable, self.r, error=True, bins=bins)
        else:
            x, y, ey = binned_curve(
                variable[~self.bads], self.r[~self.bads], error=True, bins=bins
            )
        goods = ey > 0
        p, err = polyfit(x[goods], y[goods], deg=deg, w=1.0 / ey[goods] ** 2)
        return p.reshape(-1)

    def ndof(self):
        if self.bads is None:
            Nmeas = self.A.shape[0]
        else:
            Nmeas = float(np.sum(~self.bads))
        npars = self.A.shape[1]
        return Nmeas - npars

    def reweight(self, variable, nan_exception=False, **keys):
        Nmeas = float(np.sum(~self.bads))
        npars = self.A.shape[1]

        w_model = self.weight_model(variable, **keys)
        if self.verbose:
            logger.debug(
                (
                    "Warping weights according to a polynomial fit of residuals: "
                    + str(w_model)
                )
            )
        weights = np.sqrt(1 - npars / Nmeas) / np.sqrt(np.polyval(w_model, variable))
        if nan_exception:
            weights[np.isnan(weights)] = np.nanmax(weights)

        self.A = self.A.tocoo()
        self.A.data *= weights[self.A.row]
        self.y *= weights
        self.A = self.A.tocsr()

        # Redo the numerical factorisation
        self.system.refact(self.A, self.A.T * self.y)
        self.bads = None
        return weights


class LinearModel:
    r"""Facilitate composite descriptions of linear problems of the kind:
    $$y = A_0 x_0 + A_1 x_1 + ... + A_J x_J + n$$,

    that can be subject to linear constraints of the kind:
    $$B x = d$$

    1) Each of the $A_j$ are provided in sparse format through the triplet
    rows, cols and vals. LinearModels then define the concatenation
    operator + that will handle the concatenation of matrices.

    2) Provided that the columns of the concatenated matrix $A = \left[ A_0
    A_1 ... A_J]$ remain linearly independant, the result can be passed through
    RobustLinearSolver.

    3) LinearModel handles an instance of fitparameters and makes it
    easy to fix or free a subset of parameters.

    4) Not Implemented: we may provide least-square non-trivial linear
    equality constraints at some point. This is in wait for the wrapping
    of spqr.
    """

    def __init__(self, rows, cols, vals, struct=None, name=None, valid=None):
        self.rows = np.asarray(rows)
        self.cols = np.asarray(cols)
        self.vals = np.asarray(vals)
        self._valid = valid
        self.struct = struct
        if self.struct is None:
            if name is None:
                self.struct = [self.cols.max() + 1]
            else:
                self.struct = [(name, self.cols.max() + 1)]
        self.params = FitParameters(self.struct)

    def __add__(self, m):
        # Concatenate all the pieces of the matrix
        def none_to_ones(val, length):
            if val is None:
                return np.full(length, True, dtype=bool)
            else:
                return val

        if m._valid is not None or self._valid is not None:
            valid = np.hstack(
                [
                    none_to_ones(self._valid, len(self.rows)),
                    none_to_ones(m._valid, len(m.rows)),
                ]
            )
        else:
            valid = None
        return LinearModel(
            np.hstack([self.rows, m.rows]),
            np.hstack([self.cols, m.cols + self.cols.max() + 1]),
            np.hstack([self.vals, m.vals]),
            struct=self.struct + m.struct,
            valid=valid,
        )

    def getmat(self):
        """Lazy matrix construction"""
        if not hasattr(self, "A"):
            if self._valid is None:
                self.A = sparse.coo_matrix((self.vals, (self.rows, self.cols))).tocsc()
            else:
                self.A = sparse.coo_matrix(
                    (
                        self.vals[self._valid],
                        (self.rows[self._valid], self.cols[self._valid]),
                    )
                ).tocsc()
        return self.A

    def __call__(self, p=None):
        self.getmat()
        if p is None:
            return self.A * self.params.full
        else:
            return self.A * p

    def get_free_mat(self):
        return self.getmat()[:, np.nonzero(self.params._free)[0]]

    def get_free_y(self, y):
        if not self.params._free.all():
            # I don't understand why we subtract A[:,where the param
            # is free] * the parameter value where the param is not
            # free
            return (
                y
                - self.getmat()[:, np.nonzero(~self.params._free)[0]]
                * self.params.full[~self.params._free]
            )
        else:
            # We need a copy, otherwise self.y is a reference to the
            # initial data, that will be changed by a reweighting
            # return y
            return copy.copy(y)


def linear_func(*args, **keys):
    N = len(args[0])
    for arg in args:
        assert len(arg) == N
    data = np.hstack(args)
    cols = np.repeat(np.arange(len(args)), N)
    rows = np.tile(np.arange(N), len(args))
    return LinearModel(rows, cols, data, **keys)


def indic(col_index, val=1.0, but=None, **keys):
    """Construct a sparse matrix A such that the model:

    y = x(i) * val

    with x a parameter constant for all measures sharing the same
    i can be written in matrix form as:

    Y = A X

    Parameters
    ----------
    col_index: array of int
               the indice of the bin to which each measurement belongs
    val: array of float
         the multiplicative factor for each measurement
    but: int
         if provided x(but) is withdrawn from the fit (the
         corresponding A matrix columns is suppressed)

    Exemple:
    --------

    Let us assume that we have noisy measurements of the flux of a bunch
    of stars uniquely identified by an integer 0 <= star_id < Nstar. We
    form the vector star_id such that for each measurement Y[i] for 0 <=
    i < Nmeas, star_id[i] gives the id of the corresponding star.

    The matrix implementing the data model:
    Y = A * star_flux + n

    can be obtained as
    rows, cols, vals = indic(star_id)
    A = scipy.sparse.coo_matrix((vals, (rows, cols)))
    """
    cols = col_index.copy()
    rows = np.arange(len(cols))
    data = np.ones(len(cols)) * val
    if but is not None:
        index = cols != but
        cols[cols > but] -= 1
        cols = cols[index]
        rows = rows[index]
        data = data[index]
    return LinearModel(rows, cols, data, **keys)


class TestCase:
    """Construct a fake y=Ax+n sparse linear system

    Which has a quite specific structure:
    - G global parameters
    - L local parameters
    - M measurements
    ie:
    - x has shape (G + L, 1)
    - A has shape (M, G + L)
    - y and n have shape (M, 1)
    """

    def __init__(self, G, L, M):
        cols1 = np.random.randint(0, high=G, size=M)
        index = cols1 != 0
        cols1 = cols1[index]
        model1 = indic(cols1, name="G")
        cols2 = np.repeat(np.arange(L), M / L)
        model2 = indic(cols2, name="L")
        self.model = model1 + model2
        self.x = np.random.randn(L + G)
        self.y = self.model(self.x)
        self.flux = np.random.random(M)
        self.y += np.random.randn(M) * (np.polyval((1, 1, 1), self.flux))
        self.model.params["G"].fix(0, self.x[0])
        self.x = structarray(self.x, self.model.params._struct)


# if __name__ == "__main__":
#     testcase = TestCase(10 ** 3, 10 ** 4, 10 ** 5)
#     lm = RobustLinearSolver(testcase.model, testcase.y)
#     solution1 = lm.robust_solution(nsig=5)
#     lm.diagnostic_plot([testcase.flux])
#     lm.reweight(testcase.flux)
#     solution2 = lm.robust_solution(nsig=5)
#     lm.diagnostic_plot([testcase.flux])

#     from matplotlib.pyplot import show, hist, figure
#     figure()
#     testcase.model.params.free=solution1
#     print(((testcase.model.params['G'].full - testcase.x['G']).std()))
#     hist(testcase.model.params['G'].full - testcase.x['G'], bins=20)

#     testcase.model.params.free=solution2
#     print(((testcase.model.params['G'].full - testcase.x['G']).std()))
#     hist(testcase.model.params['G'].full - testcase.x['G'], bins=20)

#     show()



import numpy as np
import pandas


class DataProxy:
    """Provide flexible ntuple manipulation.

    In particular allows to map fields of a ntuple to arbitrary field
    names and eases the creation of index for fields.
    """

    def __init__(self, datasrc, **keys):
        """Parameters:
        -----------

        keys: provides a mapping between fields of the ntuple and
        arbitrary names. For exemple, keys mag='mmag' will create
        the view: dp.mag = nt['mmag']
        """
        # Load data
        self.nt = datasrc
        self.mapping = keys
        self._create_views()
        self._index_list = set()
        self._proxy_list = []
        self._external_field_list = []

    def __len__(self):
        return len(self.nt)

    def _create_views(self):
        # Create views
        for k, v in self.mapping.items():
            setattr(self, k, self.nt[v])

    def add_views(self, **keys):
        """A view is a mapping of a NTuple field: dp.key = nt['otherkeyname']"""
        for k, v in keys.items():
            setattr(self, k, self.nt[v])
        self.mapping.update(keys)

    def _update_proxys(self):
        for k, v in self._proxy_list:
            setattr(self, k, v(self))

    def add_proxy(self, name, f):
        self._proxy_list.append((name, f))
        setattr(self, name, f(self))

    def add_partial(self, name, data, index):
        # self._partial_list.append((name,data,index))
        s = np.zeros(index.max() + 1)
        s[index.astype("int")] = data
        setattr(self, name, s)

    def add_field(self, name, data):
        """Add external data sharing the same number of records.

        There is no mem copy so this is more memory efficient than
        just catenating the field to the existing NTuple."""
        if name not in self._external_field_list:
            self._external_field_list.append(name)
        if isinstance(self.nt, (pandas.DataFrame, pandas.Series)):
            setattr(self, name, pandas.Series(data))
        else:
            setattr(self, name, data)

    def _update_external_fields(self, condition):
        for name in self._external_field_list:
            setattr(self, name, getattr(self, name)[condition])

    def compress(self, condition):
        self.nt = self.nt[condition].squeeze()
        self._create_views()
        self._update_external_fields(condition)
        self._update_proxys()
        self._update_index(condition)

    def make_index(self, fieldname, intmap=False):
        """Build an index of the values taken by field fieldname.

        Exemples:
        ---------
        Let assume that self.mjd = [12,14,14]
        calling self.make_index('mjd') will create:
        self.mjd_set = [12,14]
        self.mjd_map = {12:1, 14:2}
        self.mjd_index = [1,2,2]
        """
        self.update_index(fieldname, intmap)
        self._index_list.add((fieldname, intmap))

    def update_index(self, fieldname, intmap=False):
        if intmap:
            a = getattr(self, fieldname).astype("int")
            s = set(a)
            s = list(s)
            s.sort()
            miv = a.min()
            mav = a.max()
            m = np.ones(mav - miv + 1, dtype="int") * -1
            s = np.fromiter(s, dtype="int")
            for i, e in enumerate(s):
                m[e - miv] = i
            setattr(self, fieldname + "_set", s)
            setattr(self, fieldname + "_map", m)
            setattr(self, fieldname + "_index", m[a - miv])
        else:
            a = getattr(self, fieldname)
            s = set(a)
            s = np.array(list(s))  # np.fromiter(s, dtype=type(a[0]))
            m = dict(list(zip(s, list(range(len(s))))))
            setattr(self, fieldname + "_set", s)
            setattr(self, fieldname + "_map", m)
            i = np.fromiter((m[e] for e in a), "int")
            setattr(self, fieldname + "_index", i)

    def _update_index(self, condition):
        for fieldname, intmap in self._index_list:
            self.update_index(fieldname, intmap)


# for ind in (getattr(self, i) for i in self._index_list):
#    ind = ind[condition]



"""
Least-square fits of non-linear models with up to O(10^3 - 10^4) pars.

This module contains utility classes to perform least square fits of
(potentially large) non-linear models to experimental data.  The two
main features implemented are

- `Chi2`: a class that take a model and a dataset as arguments and
  computes the value, gradient and Hessian of the chi2 function. The
  computation tries to be efficient by using sparse matrices. This
  yield very significant speed-ups if the problem is sparse.

- `fmin_ncg`: a rewrite of scipy.optimize.fmin_ncg that accepts sparse
Jacobian and Hessian and take specific care to the factorize the
computation of the chi2 value and its derivatives.

A chi2 object may still be passed to any minimizer, notably those in
``scipy.optimize''.
"""

import logging

import numpy as np
import scipy.optimize
import scipy.sparse.linalg
import selinv
from sksparse.cholmod import cholesky, cholesky_AAt

_epsilon = 1.4901161193847656e-08


class Chi2(object):
    def __init__(self, model, x, y, w=None, s=None):
        r"""Chi2 funtion of a non-linear model of the form:

        $$\chi^2 = R^t R$$
        with:
        $$R = w (y - model(x, p))$$

        Attention: w is the square root of the inverse of the covariance
        matrix of y. At the moment we only support diagonal covariance
        matrices, specified by their diagonal as a 1d-array because I
        made specific optimizations for this case.

        Parameters:
        -----------
        model: callable
               call signature must be model(x, p, jac) and the return value must be either
               - model(x, p) if jac=False
               - (model(x,p), dmodel / dp) if jac is True, and dmodel/dp must be a sparse coo matrix
        x: x is passed directly as an argument to model
        y: 1d array
           The measurement. If si or w are given the data will be weighted
        s: array or None (broadcastable to the shape of y) expected to be the 1-sigma uncertainty of the measurements
        w: array or None (broadcastable to the shape of y) expected to be the inverse of s.
        """
        self.model = model
        self.xi = x
        self.yi = y
        assert w is None or s is None, "Cannot provide w and s at the same time"

        self.w = None
        if w is not None:
            self.w = w
        if s is not None:
            self.w = 1.0 / s

    def __call__(self, p, grad=False, hess=False, goods=None):
        # Compute the model and the Jacobian if needed
        if grad or hess:
            m, J = self.model(self.xi, p, jac=True)
            if self.w is not None:
                J.data *= self.w[J.row]
            J = J.tocsr()
            if goods is not None:
                J = J[goods, :]
        else:
            m = self.model(self.xi, p, jac=False)
        # Weight the residuals if weights are available
        wres = self.yi - m
        if self.w is not None:
            wres *= self.w
        if goods is not None:
            wres = wres[goods]

        # Chi2
        ret = [wres.T.dot(wres)]
        # print ret
        if not (grad or hess):
            return ret[0]
        if grad:
            ret.append(-2.0 * J.T.dot(wres))
        if hess:
            ret.append(2.0 * J.T * J)

        return ret

    def get_res(self, p, goods=None):
        m = self.model(self.xi, p, jac=False)
        wres = self.yi - m
        if self.w is not None:
            wres *= self.w
        if goods is not None:
            return wres[goods]
        else:
            return wres

    def gradient(self, p, **keys):
        return self(p, grad=True, **keys)[1]

    def hessian(self, p, **keys):
        return self(p, hess=True, **keys)[1]

    def densehessian(self, p, **keys):
        return self(p, hess=True, **keys)[1].todense()

    def get_cov(self, p, **keys):
        H = self.hessian(p, **keys)
        C = scipy.sparse.linalg.inv(H)
        return 2 * C

    def get_diag_cov(self, p, **keys):
        H = self.hessian(p, **keys)
        if hasattr(self, "factor"):
            return 2 * selinv.selinv(H, self.factor.P())
        else:
            return 2 * selinv.selinv(H)

    def minimize(
        self,
        p0,
        algorithm="fmin_ncg",
        goods=None,
        maxiter=2e3,
        fallback=True,
        verbose=True,
        **keys,
    ):
        if algorithm == "fmin_ncg":
            return fmin_ncg(self, p0, self.gradient, fhess=self.hessian, **keys)
        elif algorithm == "simplex":
            return scipy.optimize.fmin(self, p0, **keys)
        elif algorithm == "slsqp":
            return scipy.optimize.fmin_slsqp(self, p0, fprime=self.gradient, **keys)
        elif algorithm == "gauss_newton_homebrew":
            # Gauss-Newton with a confidence region check to switch to
            # steepest decent in case of trouble. Tipycally fails to
            # converge in high-residual cases. Try Newton or
            # quasi-Newton in this cases.
            raise_exception = keys.pop("raise", False)
            x = p0.copy()
            epsilon4 = 0.5
            it = 0
            # if "maxiter" in keys.keys():
            #     maxiter = keys["maxiter"]
            # else:
            #     maxiter = 2e3
            # if "fallback" in keys.keys():
            #     fallback = keys["fallback"]
            # else:
            #     fallback = True

            fup = 2.0
            while True:
                m, J = self.model(self.xi, x, jac=True)
                if self.w is not None:
                    J.data *= self.w[J.row]
                J = J.tocsr()
                if goods is not None:
                    J = J[goods, :]
                    m = m[goods]
                    yi = self.yi[goods]
                else:
                    yi = self.yi
                wres = yi - m
                if self.w is not None:
                    if goods is not None:
                        wres *= self.w[goods]
                    else:
                        wres *= self.w
                chi2 = wres.T.dot(wres)  # Why not call Chi2.__call__ ?
                g = J.T.dot(wres)  # SZF why not -2 * this ?
                if it == 0:
                    self.factor = cholesky_AAt(J.T)
                else:
                    self.factor.cholesky_AAt_inplace(J.T)
                deltax = self.factor(g).squeeze()
                chi2new = self(x + deltax, grad=False, hess=False, goods=goods)
                actual_red = chi2 - chi2new
                # print np.linalg.cond((J.T * J).todense())
                # For a linear f:
                pred_red = np.dot(deltax, g)
                rho = actual_red / pred_red
                dof = len(yi) - len(x)
                if verbose:
                    print(
                        (
                            "chi2=%.2f (dof=%.2f) -> %.2f (%.2f / %.2f = %g), |g|=%g"
                            % (
                                chi2,
                                dof,
                                chi2new,
                                actual_red,
                                pred_red,
                                rho,
                                np.sqrt(np.dot(g, g)),
                            )
                        )
                    )
                if (
                    not (fallback) or rho > 0
                ):  # accept the Gauss-Newton step even if it is rather poor
                    x += deltax
                    # Convergence criterion
                    if actual_red < 0.01:  # and pred_red < 0.01 and rho > 0.9:
                        break
                elif fallback and rho <= 0:  # attempt to save the meuble
                    if verbose:
                        print("bad step: trying steepest decent")

                    def c(mu):
                        return self(x + mu * g, grad=False, hess=False, goods=goods)

                    mu, chi2new, niter, funcalls = scipy.optimize.brent(
                        c, brack=(0, 1), tol=1e-2, full_output=True
                    )
                    if verbose:
                        print(
                            (
                                "Brent found new minimum in %d iter: x + %e * g -> %.2f"
                                % (niter, mu, chi2new)
                            )
                        )
                    if mu < 0:
                        if raise_exception:
                            mu = np.linspace(-1e-4, 1e-4, 20)
                            c = [
                                self(x + _mu * g, grad=False, hess=False) for _mu in mu
                            ]
                            import matplotlib.pyplot as plt

                            plt.plot(mu, c)
                            plt.show()
                            raise ValueError("Dear Lord, help us out of this merdier")
                        else:
                            pass
                    x += mu * g

                    # print "bad step"
                    # d = np.bincount(J.indices, J.data * J.data)
                    # #deltax = g / d
                    # H = J.T * J
                    # d = scipy.sparse.diags(d, 0).tocsc()
                    # nu = 0.1
                    # while (chi2new > chi2):
                    #     self.factor.cholesky_inplace(J.T*J + nu * d)
                    #     #self.factor.cholesky_AAt_inplace(J.T, nu)
                    #     deltax = self.factor(g).squeeze()
                    #     chi2new = self(x + deltax, grad=False, hess=False)
                    #     print chi2new
                    #     nu *= 2
                    #     if nu > 1e2:
                    #         # Holly shit we are in a bad shape
                    #         mu = np.linspace(0,1e-4,10)
                    #         c = [self(x + _mu * g, grad=False, hess=False) for _mu in mu]
                    #         import matplotlib.pyplot as plt
                    #         plt.plot(mu, c)
                    #         plt.show()
                    #         raise ValueError("Dear Lord, help us out of this merdier")
                    #         #x2 = scipy.optimize.anneal(self, x)
                    #         #deltax = x2 - x
                    # print "smoothing param: %g (chi2: %.2f -> %.2f)" % (nu, chi2, chi2new)
                    # x += deltax
                    # #x = scipy.optimize.fmin(self, x)

                it = it + 1
                if it > maxiter:
                    raise ValueError("Grosse fatigue")
                    if verbose:
                        print("Reach maximum number of iteration")
                    break
            chi2 = self(x, goods=goods)
            dof = len(yi) - len(x)
            if verbose:
                print(
                    (
                        "Converged in %d iterations: chi2=%g, d.o.f.=%d, chi2/dof=%.3f"
                        % (it, chi2, dof, chi2 / dof)
                    )
                )
            if keys.pop("full_output", False):
                return x, chi2, dof, self.factor
            else:
                return x

        elif algorithm == "newton_homebrew":
            # Gauss-Newton with a switch to full newton when steps fail. Requires the model to be able to compute
            # the second order part the hessian given the
            # residuals.
            raise_exception = keys.pop("raise", False)
            x = p0.copy()
            it = 0
            while True:
                m, J = self.model(self.xi, x, jac=True)
                if self.w is not None:
                    J.data *= self.w[J.row]
                J = J.tocsr()
                if goods is not None:
                    J = J[goods, :]
                    m = m[goods]
                    yi = self.yi[goods]
                else:
                    yi = self.yi
                wres = yi - m
                if self.w is not None:
                    if goods is not None:
                        wres *= self.w[goods]
                    else:
                        wres *= self.w
                chi2 = wres.T.dot(wres)
                g = J.T.dot(wres)
                H = J.T * J
                if it == 0:
                    self.factor = cholesky(H)
                else:
                    self.factor.cholesky_inplace(H)
                deltax = self.factor(g).squeeze()
                chi2new = self(x + deltax, grad=False, hess=False, goods=goods)
                actual_red = chi2 - chi2new
                # print np.linalg.cond((J.T * J).todense())
                # For a linear f:
                pred_red = np.dot(deltax, g)
                rho = actual_red / pred_red
                dof = len(yi) - len(x)
                print(
                    (
                        "%.2f (%.2f) -> %.2f (%.2f / %.2f = %g), |g|=%g"
                        % (
                            chi2,
                            dof,
                            chi2new,
                            actual_red,
                            pred_red,
                            rho,
                            np.sqrt(np.dot(g, g)),
                        )
                    )
                )
                # accept the Gauss-Newton step even if it is rather poor
                if rho >= 0:
                    x += deltax
                # Convergence criterion
                if actual_red < 0.01 and pred_red < 0.01:  # and rho >= 0:
                    break
                if 0:  # attempt a full newton step
                    Q = self.model.hess(x, wres)
                    H = H + Q
                    self.factor = cholesky(H)
                    deltax = self.factor(g).squeeze()
                    R = g - H * deltax
                    delta2x = self.factor(R)
                    print((abs(delta2x / deltax).max()))
                    chi2new = self(x + deltax, grad=False, hess=False)
                    print(chi2new)
                    chi2new = self(x + deltax + delta2x, grad=False, hess=False)
                    print(chi2new)
                    actual_red = chi2 - chi2new
                    pred_red = np.dot(deltax, g)
                    rho = actual_red / pred_red
                    print(
                        (
                            "%.2f (%.2f) -> %.2f (%.2f / %.2f = %g), |g|=%g"
                            % (
                                chi2,
                                dof,
                                chi2new,
                                actual_red,
                                pred_red,
                                rho,
                                np.sqrt(np.dot(g, g)),
                            )
                        )
                    )
                    if rho < 0:
                        print("Newton step failed. Stopping")
                        break
                    else:
                        x += deltax
                        if actual_red < 0.01 and pred_red < 0.01 and rho > 0.9:
                            print("optimization terminated succesfully")
                            break

                it = it + 1
                if it > 2e3:
                    raise ValueError("Grosse fatigue")
            chi2 = self(x, goods=goods)
            dof = len(yi) - len(x)
            print(
                (
                    "Converged in %d iterations: chi2=%g, d.o.f.=%d, chi2/dof=%.3f"
                    % (it, chi2, dof, chi2 / dof)
                )
            )
            if keys.pop("full_output", False):
                return x, chi2, dof, self.factor
            else:
                return x

        elif algorithm == "gauss_newton_dense":
            # Gauss-Newton with a confidence region check to switch to
            # steepest decent in case of trouble. Tipycally fails to
            # converge in high-residual cases. Try Newton or
            # quasi-Newton in this cases.
            raise_exception = keys.pop("raise", False)
            x = p0.copy()
            # epsilon4 = 0.5
            it = 0
            # if "maxiter" in keys.keys():
            #     maxiter = keys["maxiter"]
            # else:
            #     maxiter = 2e3
            # if "fallback" in keys.keys():
            #     fallback = keys["fallback"]
            # else:
            #     fallback = True

            # fup = 2.0
            while True:
                m, J = self.model(self.xi, x, jac=True)
                if self.w is not None:
                    J.data *= self.w[J.row]
                J = J.tocsr()
                if goods is not None:
                    J = J[goods, :]
                    m = m[goods]
                    yi = self.yi[goods]
                else:
                    yi = self.yi
                wres = yi - m
                if self.w is not None:
                    if goods is not None:
                        wres *= self.w[goods]
                    else:
                        wres *= self.w
                chi2 = wres.T.dot(wres)
                g = J.T.dot(wres)
                C = (J.T * J).todense()
                self.factor = scipy.linalg.cho_factor(C)
                deltax = scipy.linalg.cho_solve(self.factor, g).squeeze()
                chi2new = self(x + deltax, grad=False, hess=False, goods=goods)
                actual_red = chi2 - chi2new
                # print np.linalg.cond((J.T * J).todense())
                # For a linear f:
                pred_red = np.dot(deltax, g)
                rho = actual_red / pred_red
                dof = len(yi) - len(x)
                if verbose:
                    print(
                        (
                            "chi2=%.2f (dof=%.2f) -> %.2f (%.2f / %.2f = %g), |g|=%g"
                            % (
                                chi2,
                                dof,
                                chi2new,
                                actual_red,
                                pred_red,
                                rho,
                                np.sqrt(np.dot(g, g)),
                            )
                        )
                    )
                if (
                    not (fallback) or rho > 0
                ):  # accept the Gauss-Newton step even if it is rather poor
                    x += deltax
                    # Convergence criterion
                    if actual_red < 0.01:  # and pred_red < 0.01 and rho > 0.9:
                        break
                else:
                    print("bad step, giving up")
                    break
                it = it + 1
                if it > maxiter:
                    raise ValueError("Grosse fatigue")
                    if verbose:
                        print("Reach maximum number of iteration")
                    break
            chi2 = self(x, goods=goods)
            dof = len(yi) - len(x)
            if verbose:
                print(
                    (
                        "Converged in %d iterations: chi2=%g, d.o.f.=%d, chi2/dof=%.3f"
                        % (it, chi2, dof, chi2 / dof)
                    )
                )
            if keys.pop("full_output", False):
                return x, chi2, dof, self.factor
            else:
                return x
        else:
            raise ValueError("Unknown algorithm")


# def lsqfit_gauss_newton(y, model, p0, eq_constraints=None, weights=None, **keys):
#    """ Gauss-Newton with a confidence region check to switch to
#    steepest decent in case of trouble. Tipycally fails to
#    converge in high-residual cases. Try Newton or
#    quasi-Newton in this cases.
#    """
#    raise_exception = keys.pop('raise', False)
#    x = p0.copy()
#    while True:
#        m, J = model(0, x, jac=True)
#        if weights is not None:
#            J.data *= weights[J.row]
#        J = J.tocsr()
#        wres = y - m
#        if weights is not None:
#            wres *= weights
#        chi2 = wres.T.dot(wres)
#        g = J.T.dot(wres)
#        if eq_constraints is not None:
#            c, A = eq_constraints(x, jac=True)
#            H = scipy.sparse.bmat([[J.T*J, A],[A.T, None]])
#            B = np.hstack([g, -c])
#            lu = sparse.linalg.splu(H)
#            dxl = lu.solve(B)
#            dx = dxl[:-len(c)]
#            lagrange = dxl[-len(c):]
#            if it == 0:
#                    self.factor = cholesky_AAt(J.T)
#                else:
#                    self.factor.cholesky_AAt_inplace(J.T)
#                deltax = self.factor(g).squeeze()
#                chi2new = self(x + deltax, grad=False, hess=False)
#                actual_red = chi2 - chi2new
#                #print np.linalg.cond((J.T * J).todense())
#                # For a linear f:
#                pred_red = (np.dot(deltax, g))
#                rho = actual_red / pred_red
#                dof = len(self.yi) - len(x)
#                print "%.2f (%.2f) -> %.2f (%.2f / %.2f = %g), |g|=%g" % (chi2, dof, chi2new, actual_red, pred_red, rho, np.sqrt(np.dot(g,g)))
#                if rho > 0: # accept the Gauss-Newton step even if it is rather poor
#                    x += deltax
#                    # Convergence criterion
#                    if actual_red < 0.01 and pred_red < 0.01 and rho > 0.9:
#                        break
#                else:  # attempt to save the meuble
#                    print "bad step: trying steepest decent"
#                    def c(mu):
#                        return self(x + mu * g, grad=False, hess=False)
#                    mu, chi2new, niter, funcalls = scipy.optimize.brent(c, brack=(0,1), tol=1e-2, full_output=True)
#                    print "Brent found new minimum in %d iter: x + %e * g -> %.2f" % (niter, mu, chi2new)
#                    if mu < 0:
#                        if raise_exception:
#                            mu = np.linspace(-1e-4,1e-4,20)
#                            c = [self(x + _mu * g, grad=False, hess=False) for _mu in mu]
#                            import matplotlib.pyplot as plt
#                            plt.plot(mu, c)
#                            plt.show()
#                            raise ValueError("Dear Lord, help us out of this merdier")
#                        else:
#                            pass
#                    x += mu * g
#                    """
#                    print "bad step"
#                    d = np.bincount(J.indices, J.data * J.data)
#                    #deltax = g / d
#                    H = J.T * J
#                    d = scipy.sparse.diags(d, 0).tocsc()
#                    nu = 0.1
#                    while (chi2new > chi2):
#                        self.factor.cholesky_inplace(J.T*J + nu * d)
#                        #self.factor.cholesky_AAt_inplace(J.T, nu)
#                        deltax = self.factor(g).squeeze()
#                        chi2new = self(x + deltax, grad=False, hess=False)
#                        print chi2new
#                        nu *= 2
#                        if nu > 1e2:
#                            # Holly shit we are in a bad shape
#                            mu = np.linspace(0,1e-4,10)
#                            c = [self(x + _mu * g, grad=False, hess=False) for _mu in mu]
#                            import matplotlib.pyplot as plt
#                            plt.plot(mu, c)
#                            plt.show()
#                            raise ValueError("Dear Lord, help us out of this merdier")
#                            #x2 = scipy.optimize.anneal(self, x)
#                            #deltax = x2 - x
#                    print "smoothing param: %g (chi2: %.2f -> %.2f)" % (nu, chi2, chi2new)
#                    x += deltax
#                    #x = scipy.optimize.fmin(self, x)
#                    """
#                it = it + 1
#                if it > 2e3:
#                    raise ValueError("Grosse fatigue")
#            chi2 = self(x)
#            dof = len(self.yi) - len(x)
#            print "Converged in %d iterations: chi2=%g, d.o.f.=%d, chi2/dof=%.3f" % (it, chi2, dof, chi2/dof)
#            if keys.pop('full_output', False):
#                return x, chi2, dof, self.factor
#            else:
#                return x
#
def fmin_ncg(
    f,
    x0,
    fprime,
    fhess_p=None,
    fhess=None,
    args=(),
    avextol=1e-5,
    epsilon=_epsilon,
    maxiter=None,
    full_output=0,
    disp=1,
    retall=0,
    callback=None,
):
    """
    Unconstrained minimization of a function using the Newton-CG method.

    This is a patched version of the scipy.optimize function accepting
    sparse Hessian matrices.

    Parameters
    ----------
    f : callable ``f(x, *args)``
        Objective function to be minimized.
    x0 : ndarray
        Initial guess.
    fprime : callable ``f'(x, *args)``
        Gradient of f.
    fhess_p : callable ``fhess_p(x, p, *args)``, optional
        Function which computes the Hessian of f times an
        arbitrary vector, p.
    fhess : callable ``fhess(x, *args)``, optional
        Function to compute the Hessian matrix of f.
    args : tuple, optional
        Extra arguments passed to f, fprime, fhess_p, and fhess
        (the same set of extra arguments is supplied to all of
        these functions).
    epsilon : float or ndarray, optional
        If fhess is approximated, use this value for the step size.
    callback : callable, optional
        An optional user-supplied function which is called after
        each iteration.  Called as callback(xk), where xk is the
        current parameter vector.
    avextol : float, optional
        Convergence is assumed when the average relative error in
        the minimizer falls below this amount.
    maxiter : int, optional
        Maximum number of iterations to perform.
    full_output : bool, optional
        If True, return the optional outputs.
    disp : bool, optional
        If True, print convergence message.
    retall : bool, optional
        If True, return a list of results at each iteration.

    Returns
    -------
    xopt : ndarray
        Parameters which minimize f, i.e. ``f(xopt) == fopt``.
    fopt : float
        Value of the function at xopt, i.e. ``fopt = f(xopt)``.
    fcalls : int
        Number of function calls made.
    gcalls : int
        Number of gradient calls made.
    hcalls : int
        Number of hessian calls made.
    warnflag : int
        Warnings generated by the algorithm.
        1 : Maximum number of iterations exceeded.
    allvecs : list
        The result at each iteration, if retall is True (see below).

    See also
    --------
    minimize: Interface to minimization algorithms for multivariate
        functions. See the 'Newton-CG' `method` in particular.

    Notes
    -----
    Only one of `fhess_p` or `fhess` need to be given.  If `fhess`
    is provided, then `fhess_p` will be ignored.  If neither `fhess`
    nor `fhess_p` is provided, then the hessian product will be
    approximated using finite differences on `fprime`. `fhess_p`
    must compute the hessian times an arbitrary vector. If it is not
    given, finite-differences on `fprime` are used to compute
    it.

    Newton-CG methods are also called truncated Newton methods. This
    function differs from scipy.optimize.fmin_tnc because

    1. scipy.optimize.fmin_ncg is written purely in python using numpy
        and scipy while scipy.optimize.fmin_tnc calls a C function.
    2. scipy.optimize.fmin_ncg is only for unconstrained minimization
        while scipy.optimize.fmin_tnc is for unconstrained minimization
        or box constrained minimization. (Box constraints give
        lower and upper bounds for each variable separately.)

    References
    ----------
    Wright & Nocedal, 'Numerical Optimization', 1999, pg. 140.

    """
    opts = {
        "xtol": avextol,
        "eps": epsilon,
        "maxiter": maxiter,
        "disp": disp,
        "return_all": retall,
    }

    res = _minimize_newtoncg(
        f, x0, args, fprime, fhess, fhess_p, callback=callback, **opts
    )

    if full_output:
        retlist = (
            res["x"],
            res["fun"],
            res["nfev"],
            res["njev"],
            res["nhev"],
            res["status"],
        )
        if retall:
            retlist += (res["allvecs"],)
        return retlist
    else:
        if retall:
            return res["x"], res["allvecs"]
        else:
            return res["x"]


def _minimize_newtoncg(
    fun,
    x0,
    args=(),
    jac=None,
    hess=None,
    hessp=None,
    callback=None,
    xtol=1e-5,
    eps=_epsilon,
    maxiter=None,
    disp=False,
    return_all=False,
    **unknown_options,
):
    """
    Minimization of scalar function of one or more variables using the
    Newton-CG algorithm.

    This is a patched version of the scipy.optimize function accepting
    sparse hessian.

    Options for the Newton-CG algorithm are:
        disp : bool
            Set to True to print convergence messages.
        xtol : float
            Average relative error in solution `xopt` acceptable for
            convergence.
        maxiter : int
            Maximum number of iterations to perform.
        eps : float or ndarray
            If `jac` is approximated, use this value for the step size.

    This function is called by the `minimize` function with
    `method=Newton-CG`. It is not supposed to be called directly.

    Also note that the `jac` parameter (Jacobian) is required.
    """
    _check_unknown_options(unknown_options)
    if jac is None:
        raise ValueError("Jacobian is required for Newton-CG method")
    f = fun
    fprime = jac
    fhess_p = hessp
    fhess = hess
    avextol = xtol
    epsilon = eps
    retall = return_all

    x0 = np.asarray(x0).flatten()
    fcalls, f = _wrap_function(f, args)
    gcalls, fprime = _wrap_function(fprime, args)
    hcalls = 0
    if maxiter is None:
        maxiter = len(x0) * 200

    xtol = len(x0) * avextol
    update = [2 * xtol]
    xk = x0
    if retall:
        allvecs = [xk]
    k = 0
    old_fval = f(x0)
    old_old_fval = None
    float64eps = np.finfo(np.float64).eps
    warnflag = 0
    while (np.add.reduce(np.abs(update)) > xtol) and (k < maxiter):
        # Compute a search direction pk by applying the CG method to
        #  del2 f(xk) p = - grad f(xk) starting from 0.

        # b = -fprime(xk)
        c2, b, A = fun(xk, grad=True, hess=True)
        b = -b
        hcalls = hcalls + 1
        # gcalls = gcalls + 1
        maggrad = np.add.reduce(np.abs(b))
        eta = np.min([0.5, np.sqrt(maggrad)])
        termcond = eta * maggrad
        xsupi = np.zeros(len(x0), dtype=x0.dtype)
        ri = -b
        psupi = -ri
        i = 0
        dri0 = np.dot(ri, ri)

        # if fhess is not None:             # you want to compute hessian once.
        #    A = fhess(*(xk,) + args)
        #    hcalls = hcalls + 1

        while np.add.reduce(np.abs(ri)) > termcond:
            # if fhess is None:
            #    if fhess_p is None:
            #        Ap = approx_fhess_p(xk, psupi, fprime, epsilon)
            #    else:
            #        Ap = fhess_p(xk, psupi, *args)
            #        hcalls = hcalls + 1
            # else:
            # Ap = np.dot(A, psupi)
            Ap = A.dot(psupi)
            # check curvature
            Ap = np.asarray(Ap).squeeze()  # get rid of matrices...
            curv = np.dot(psupi, Ap)
            if 0 <= curv <= 3 * float64eps:
                break
            elif curv < 0:
                if i > 0:
                    break
                else:
                    # fall back to steepest descent direction
                    xsupi = dri0 / (-curv) * b
                    break
            alphai = dri0 / curv
            xsupi = xsupi + alphai * psupi
            ri = ri + alphai * Ap
            dri1 = np.dot(ri, ri)
            betai = dri1 / dri0
            psupi = -ri + betai * psupi
            i = i + 1
            dri0 = dri1  # update np.dot(ri,ri) for next time.

        pk = xsupi  # search direction is solution to system.
        gfk = -b  # gradient at xk

        try:
            alphak, fc, gc, old_fval, old_old_fval, gfkp1 = _line_search_wolfe12(
                f, fprime, xk, pk, gfk, old_fval, old_old_fval
            )
        except _LineSearchError:
            # Line search failed to find a better solution.
            warnflag = 2
            break

        update = alphak * pk
        xk = xk + update  # upcast if necessary
        if callback is not None:
            callback(xk)
        if retall:
            allvecs.append(xk)
        k += 1

    fval = old_fval
    if warnflag == 2:
        msg = _status_message["pr_loss"]
        if disp:
            print(("Warning: " + msg))
            print(("         Current function value: %f" % fval))
            print(("         Iterations: %d" % k))
            print(("         Function evaluations: %d" % fcalls[0]))
            print(("         Gradient evaluations: %d" % gcalls[0]))
            print(("         Hessian evaluations: %d" % hcalls))
    elif k >= maxiter:
        warnflag = 1
        msg = _status_message["maxiter"]
        if disp:
            print(("Warning: " + msg))
            print(("         Current function value: %f" % fval))
            print(("         Iterations: %d" % k))
            print(("         Function evaluations: %d" % fcalls[0]))
            print(("         Gradient evaluations: %d" % gcalls[0]))
            print(("         Hessian evaluations: %d" % hcalls))
    else:
        msg = _status_message["success"]
        if disp:
            print(msg)
            print(("         Current function value: %f" % fval))
            print(("         Iterations: %d" % k))
            print(("         Function evaluations: %d" % fcalls[0]))
            print(("         Gradient evaluations: %d" % gcalls[0]))
            print(("         Hessian evaluations: %d" % hcalls))

    result = OptimizeResult(
        fun=fval,
        jac=gfk,
        nfev=fcalls[0],
        njev=gcalls[0],
        nhev=hcalls,
        status=warnflag,
        success=(warnflag == 0),
        message=msg,
        x=xk,
    )
    if retall:
        result["allvecs"] = allvecs
    return result


class RobustMinimizer:
    """ """

    def __init__(self, model, y, x=None, weights=None, verbose=1):
        """ """
        self.model = model
        self.x = x
        self.y = y
        self.w = weights
        self.verbose = verbose
        self._chi2 = Chi2(model, x, y, w=weights)
        N = len(self.y)
        self.bads = np.zeros(N).astype(bool)
        self.suppress = None
        self.restore = None

    def minimize(self, p0, algorithm="newton_homebrew", bads=None, **kwargs):
        """ """
        # restore the initial bads
        if bads is None:
            self.bads = np.zeros(len(self.y)).astype(bool)
        else:
            self.bads = bads

        # first minimization
        p = self._chi2.minimize(p0, goods=(~self.bads), algorithm=algorithm, **kwargs)

        niter = 1
        while self.flag_outliers(p, **kwargs):
            logging.info("nbads/ntot=%d/%d" % (self.bads.sum(), len(self.bads)))

            pp = self._chi2.minimize(
                p, goods=(~self.bads), algorithm=algorithm, **kwargs
            )
            niter += 1
            logging.info(
                "dp=%g chi2=%g"
                % (np.sum(np.abs(pp - p)), self._chi2(pp, goods=~self.bads))
            )
            p[:] = pp

        if self.verbose:
            logging.info(
                "converged after %d iterations %d/%d outliers removed"
                % (niter, self.bads.sum(), len(self.bads))
            )

        self.params = self.model.params.copy()
        self.params.free = p
        self.eparams = self.model.params.copy()
        self.eparams.free = np.sqrt(self.get_diag_cov(p))

        return p

    def check_local_param(self, p, local_param):
        """ """
        v, J = self.model(self.x, p, jac=True)
        local_cols = slice(
            self.model.params[local_param].indexof(0),
            self.model.params[local_param].indexof(-1),
        )
        J = J.tocsc()[~self.bads, local_cols]
        return J.sum(axis=0)

    def suppress_one_at_a_time(self, r, p, local_param):
        """
        For each paramter, suppress only the worst outlier

        """
        if self.suppress.sum() == 0:
            return self.suppress

        v, J = self.model(self.x, p, jac=1)
        local_cols = slice(
            self.model.params[local_param].indexof(0),
            self.model.params[local_param].indexof(-1),
        )
        J = J.tocsc()[self.suppress, local_cols].tocoo()
        J.data = r[self.suppress][J.row]

        # the line number of the largest outlier
        i_max = np.array(J.argmax(axis=0)).squeeze()
        # it may also be that the corresponding column is empty
        v_max = np.array(J.max(axis=0).todense()).squeeze()
        # the outliers we need to remove
        to_remove = i_max[i_max > 0 | ((i_max == 0) & (v_max > 0.0))]

        suppress_one = np.zeros_like(self.suppress[self.suppress])
        suppress_one[to_remove] = True
        n_before = self.suppress.sum()
        self.suppress[self.suppress] = suppress_one
        n_after = self.suppress.sum()
        logging.info("to suppress: %d -> %d" % (n_before, n_after))

    def flag_outliers(self, p, nsig=5, local_param=None):
        """flag the outliers that deviate from the mean chi2 by at least
        nsig. The standard deviation of the residuals is the square root of the chi2
        divided by the number of degrees of freedom.

        Args:
           p (array of floats): current solution (free parameters)
           nsig (int or float): cut in sigma
           local_param (array of ints): parameters

        Returns:
           None

        .. note: modifies self.bads, self.suppress, self.restore
        """
        r = self._chi2.get_res(p)
        r *= r
        if self.bads is not None:
            chi2 = self._chi2(p, goods=~self.bads)
            nmeas = (~self.bads).sum()
        else:
            chi2 = self._chi2(p)
            nmeas = len(r)

        n_free_pars = len(p)
        ndof = nmeas - n_free_pars

        # identify the outliers
        bads = r > (nsig**2 * chi2 / ndof)

        # now, who should I actually suppress & restore ?
        self.suppress = bads & ~self.bads
        self.restore = ~bads & self.bads

        # here, I should add the local params rejection
        if local_param is not None:
            self.suppress_one_at_a_time(r, p, local_param)

        n_suppressed = self.suppress.sum()
        n_restored = self.restore.sum()

        if n_suppressed == 0:
            self.suppress = None
        else:
            self.bads[self.suppress] = True

        if n_restored == 0:
            self.restore = None
        else:
            self.bads[self.restore] = False

        # update the chi2
        new_chi2 = self._chi2(p, goods=~self.bads)

        if self.verbose > 0:
            logging.info("suppressed %d outliers" % n_suppressed)
            logging.info("restored %d outliers" % n_restored)
            logging.info("chi2: %g -> %g" % (chi2, new_chi2))

        return n_restored > 0 or n_suppressed > 0

    def get_res(self, p=None, x=None):
        """
        how do we deal with the bads ?
        """
        if p is None:
            pp = self.model.params.free
        else:
            pp = p
        r = self.model(x, p=pp)
        r -= self.y
        return r

    def get_wres(self, x=None):
        """
        how do we deal with the bads ?
        """
        r = self.model(x, p=self.params.free)
        r -= self.y
        return self.weights * r

    def get_cov(self):
        return self._chi2.get_cov(self.params.free, goods=~self.bads)

    def get_block_cov(self):
        pass

    def get_diag_cov(self, p):
        return self._chi2.get_diag_cov(p, goods=~self.bads)

    def ndof(self):
        nmeas = (~self.bads).sum()
        npars = len(self.params.free)
        return nmeas - npars

    def chi2(self, p=None):
        if p:
            return self._chi2(p, goods=~self.bads)
        return self._chi2(self.params.free, goods=~self.bads)


class OptimizeResult(dict):
    """Represents the optimization result.

    Attributes
    ----------
    x : ndarray
        The solution of the optimization.
    success : bool
        Whether or not the optimizer exited successfully.
    status : int
        Termination status of the optimizer. Its value depends on the
        underlying solver. Refer to `message` for details.
    message : str
        Description of the cause of the termination.
    fun, jac, hess, hess_inv : ndarray
        Values of objective function, Jacobian, Hessian or its inverse (if
        available). The Hessians may be approximations, see the documentation
        of the function in question.
    nfev, njev, nhev : int
        Number of evaluations of the objective functions and of its
        Jacobian and Hessian.
    nit : int
        Number of iterations performed by the optimizer.
    maxcv : float
        The maximum constraint violation.

    Notes
    -----
    There may be additional attributes not listed above depending of the
    specific solver. Since this class is essentially a subclass of dict
    with attribute accessors, one can see which attributes are available
    using the `keys()` method.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if list(self.keys()):
            m = max(list(map(len, list(self.keys())))) + 1
            return "\n".join(
                [k.rjust(m) + ": " + repr(v) for k, v in list(self.items())]
            )
        else:
            return self.__class__.__name__ + "()"



"""Functions for gnomonic projections"""

# This code originates from https://gitlab.in2p3.fr/betoule/imageproc by Marc
# Betoule.

import numpy as np

deg2rad = np.pi / 180.0


def dot3(x, M):
    """return the xMx vector where x is 2xN, M is 2x2xN"""
    xMx = np.matmul(x.T[:, None, :], np.matmul(M.T, x.T[:, :, None]))
    return np.squeeze(xMx)


def inverse_gnomonic_projection(
    x_rad, y_rad, rac_rad, decc_rad, vx_rad=None, vy_rad=None, vxy_rad=None
):
    """ """
    rho = np.sqrt(x_rad**2 + y_rad**2)
    c = np.arctan(rho)
    sinc_over_rho = np.where(rho == 0, 1, np.sin(c) / rho)
    cde, sde = np.cos(decc_rad[None, :]), np.sin(decc_rad[None, :])
    gamma = rho * cde * np.cos(c) - y_rad * sde * np.sin(c)

    dec_rad = np.arcsin(np.cos(c) * sde + y_rad * sinc_over_rho * cde)
    ra_rad = rac_rad[None, :] + np.arctan2(x_rad * np.sin(c), gamma)

    if vx_rad is not None and vy_rad is not None:
        dra, ddec = inverse_gnomonic_dradec(x_rad, y_rad, rac_rad, decc_rad)
        if vxy_rad is None:
            vxy_rad = np.zeros(len(vx_rad))
        era = np.sqrt(dot3(dra, np.array([(vx_rad, vxy_rad), (vxy_rad, vy_rad)])))
        edec = np.sqrt(dot3(ddec, np.array([(vx_rad, vxy_rad), (vxy_rad, vy_rad)])))
        return ra_rad.flatten(), dec_rad.flatten(), era, edec
    else:
        return ra_rad.flatten(), dec_rad.flatten()


def inverse_gnomonic_dcenter(x_rad, y_rad, rac_rad, decc_rad):
    """x, y in rad
    return drac, ddecc in rad
    """
    rho = np.sqrt(x_rad**2 + y_rad**2)
    c = np.arctan(rho)
    # sinc_over_rho = np.where(rho==0, 1,np.sin(c)/rho)
    cde, sde = np.cos(decc_rad[None, :]), np.sin(decc_rad[None, :])
    gamma = rho * cde * np.cos(c) - y_rad * sde * np.sin(c)
    dgammaddecc = -rho * sde * np.cos(c) - y_rad * cde * np.sin(c)

    v2 = np.sqrt(1 - (np.cos(c) * sde + (y_rad * np.sin(c) * cde) / rho) ** 2)
    g2 = 1 + (x_rad * np.sin(c) / gamma) ** 2

    ddecddecc = np.cos(c) * cde - (y_rad * np.sin(c) * sde) / rho
    ddecddecc /= v2
    draddecc = -x_rad * np.sin(c) * dgammaddecc / gamma**2
    draddecc /= g2
    ddecdrac = np.zeros((len(draddecc.flatten())), dtype=float)
    draddrac = np.ones((len(draddecc.flatten())), dtype=float)
    return np.array([draddrac, draddecc.flatten()]), np.array(
        [ddecdrac, ddecddecc.flatten()]
    )


def inverse_gnomonic_dradec(x_rad, y_rad, rac_rad, decc_rad):
    """x, y in rad
    return dra, ddec in rad
    """
    rho = np.sqrt(x_rad**2 + y_rad**2)
    c = np.arctan(rho)
    # sinc_over_rho = np.where(rho==0, 1,np.sin(c)/rho)
    cde, sde = np.cos(decc_rad[None, :]), np.sin(decc_rad[None, :])
    gamma = rho * cde * np.cos(c) - y_rad * sde * np.sin(c)
    drhodx, drhody = x_rad / rho, y_rad / rho
    dirhodx, dirhody = -x_rad / rho**3, -y_rad / rho**3
    dcdx, dcdy = 1 / (1 + rho**2) * drhodx, 1 / (1 + rho**2) * drhody
    v2 = np.sqrt(1 - (np.cos(c) * sde + (y_rad * np.sin(c) * cde) / rho) ** 2)

    ddecdx = -np.sin(c) * sde * dcdx
    ddecdx += (y_rad * np.cos(c) * cde / rho) * dcdx
    ddecdx += y_rad * np.sin(c) * cde * dirhodx
    ddecdy = -np.sin(c) * sde * dcdy
    ddecdy += cde / rho * (y_rad * np.cos(c) * dcdy + np.sin(c))
    ddecdy += y_rad * np.sin(c) * cde * dirhody
    ddec = np.array([(ddecdx / v2).flatten(), (ddecdy / v2).flatten()])

    dgammadx = drhodx * cde * np.cos(c)
    dgammadx -= rho * cde * np.sin(c) * dcdx
    dgammadx -= y_rad * sde * np.cos(c) * dcdx
    dgammady = drhody * cde * np.cos(c)
    dgammady -= rho * cde * np.sin(c) * dcdy
    dgammady -= y_rad * sde * np.cos(c) * dcdy
    dgammady -= sde * np.sin(c)
    g2 = 1 + (x_rad * np.sin(c) / gamma) ** 2

    dradx = (np.sin(c) + dcdx * x_rad * np.cos(c)) / gamma
    dradx -= x_rad * np.sin(c) * dgammadx / (gamma**2)
    drady = x_rad * np.cos(c) * dcdy / gamma
    drady -= x_rad * np.sin(c) * dgammady / (gamma**2)
    dra = np.array([(dradx / g2).flatten(), (drady / g2).flatten()])
    return dra, ddec


def gnomonic_dxy(ra_rad, dec_rad, rac_rad, decc_rad):
    """Compute the Jacobian of the gnomonic projection w.r.t. (ra, dec).

    For each source position, evaluates the partial derivatives of the
    tangent-plane coordinates (x, y) with respect to right ascension and
    declination. The result can be used to propagate positional uncertainties
    or proper motions into the tangent plane via a linear transformation.

    Parameters
    ----------
    ra_rad : array_like
        Right ascension of the source(s), in radians.
    dec_rad : array_like
        Declination of the source(s), in radians.
    rac_rad : array_like
        Right ascension of the tangent point(s) (projection center), in radians.
    decc_rad : array_like
        Declination of the tangent point(s) (projection center), in radians.

    Returns
    -------
    dx : ndarray, shape (2, N)
        Jacobian rows for the x coordinate. ``dx[0]`` = ∂x/∂ra and
        ``dx[1]`` = ∂x/∂dec, both flattened over all (source, center) pairs.
        Units are rad / rad (dimensionless).
    dy : ndarray, shape (2, N)
        Jacobian rows for the y coordinate. ``dy[0]`` = ∂y/∂ra and
        ``dy[1]`` = ∂y/∂dec, both flattened over all (source, center) pairs.
        Units are rad / rad (dimensionless).
    """
    s, c = np.sin(decc_rad[None, :]), np.cos(decc_rad[None, :])
    c2 = np.cos(ra_rad - rac_rad[None, :])
    d = np.sin(dec_rad) * s + np.cos(dec_rad) * c * c2

    ddr = -np.cos(dec_rad) * c * np.sin(ra_rad - rac_rad[None, :])
    ddd = s * np.cos(dec_rad) - c * np.sin(dec_rad) * c2

    dxdr = (
        np.cos(dec_rad) * c2 * d
        - (np.cos(dec_rad) * np.sin(ra_rad - rac_rad[None, :])) * ddr
    )

    dxdd = -np.sin(dec_rad) * np.sin(ra_rad - rac_rad[None, :]) * d
    dxdd -= (np.cos(dec_rad) * np.sin(ra_rad - rac_rad[None, :])) * ddd
    dx = np.array([(dxdr / d**2).flatten(), (dxdd / d**2).flatten()])

    dydr = np.cos(dec_rad) * s * np.sin(ra_rad - rac_rad[None, :]) * d
    dydr -= (np.sin(dec_rad) * c - np.cos(dec_rad) * s * c2) * ddr
    dydd = (np.cos(dec_rad) * c + np.sin(dec_rad) * s * c2) * d
    dydd -= (np.sin(dec_rad) * c - np.cos(dec_rad) * s * c2) * ddd
    dy = np.array([(dydr / d**2).flatten(), (dydd / d**2).flatten()])
    return dx, dy


def gnomonic_projection(
    ra_rad, dec_rad, rac_rad, decc_rad, era_rad=None, edec_rad=None, astropy=False
):
    """Project celestial coordinates onto the tangent plane (gnomonic projection).

    Maps spherical coordinates (ra, dec) to flat tangent-plane coordinates
    (x, y) using the gnomonic projection centered at (rac, decc):

        d = sin(dec)*sin(decc) + cos(dec)*cos(decc)*cos(ra - rac)
        x = cos(dec)*sin(ra - rac) / d
        y = (sin(dec)*cos(decc) - cos(dec)*sin(decc)*cos(ra - rac)) / d

    Optionally propagates uncorrelated position uncertainties into the
    tangent plane using the projection Jacobian from `gnomonic_dxy`.

    Parameters
    ----------
    ra_rad : array_like
        Right ascension of the source(s), in radians.
    dec_rad : array_like
        Declination of the source(s), in radians.
    rac_rad : array_like
        Right ascension of the tangent point(s) (projection center), in radians.
    decc_rad : array_like
        Declination of the tangent point(s) (projection center), in radians.
    era_rad : array_like, optional
        1-sigma uncertainty on right ascension, in radians. Must be provided
        together with `edec_rad` to trigger uncertainty propagation.
        Ignored when `astropy=True`.
    edec_rad : array_like, optional
        1-sigma uncertainty on declination, in radians. Must be provided
        together with `era_rad` to trigger uncertainty propagation.
        Ignored when `astropy=True`.
    astropy : bool, optional
        If ``True``, use the sign convention consistent with astropy
        (x = cos(decc)*sin(rac - ra) / d), which reverses the x-axis
        relative to the default. Uncertainty propagation is disabled in
        this mode. Default is ``False``.

    Returns
    -------
    x : ndarray
        Tangent-plane x-coordinate, in radians.
    y : ndarray
        Tangent-plane y-coordinate, in radians.
    ex : ndarray
        1-sigma uncertainty on x, in radians. Only returned when both
        `era_rad` and `edec_rad` are provided and `astropy` is ``False``.
    ey : ndarray
        1-sigma uncertainty on y, in radians. Only returned when both
        `era_rad` and `edec_rad` are provided and `astropy` is ``False``.
    """
    if astropy:
        s, c = np.sin(dec_rad), np.cos(dec_rad)
        c2 = np.cos(rac_rad[None, :] - ra_rad)
        d = np.sin(decc_rad[None, :]) * s + np.cos(decc_rad[None, :]) * c * c2
        x = (np.cos(decc_rad[None, :]) * np.sin(rac_rad[None, :] - ra_rad)) / d
        y = (np.sin(decc_rad[None, :]) * c - np.cos(decc_rad[None, :]) * s * c2) / d
    else:
        if len(dec_rad.shape) == 2 and dec_rad.shape[1] == 4:
            de, re = decc_rad[:, None], rac_rad[:, None]
        else:
            de, re = decc_rad[None, :], rac_rad[None, :]

        s, c = np.sin(de), np.cos(de)
        c2 = np.cos(ra_rad - re)
        d = np.sin(dec_rad) * s + np.cos(dec_rad) * c * c2
        x = (np.cos(dec_rad) * np.sin(ra_rad - re)) / d
        y = (np.sin(dec_rad) * c - np.cos(dec_rad) * s * c2) / d

    if era_rad is not None and edec_rad is not None and not astropy:
        dx, dy = gnomonic_dxy(ra_rad, dec_rad, rac_rad, decc_rad)
        z = np.zeros(len(era_rad))
        ex = np.sqrt(dot3(dx, np.array([(era_rad**2, z), (z, edec_rad**2)])))
        ey = np.sqrt(dot3(dy, np.array([(era_rad**2, z), (z, edec_rad**2)])))
        return x, y, ex, ey
    else:
        return x, y  # x.flatten(), y.flatten()


def tangent_plane_projection(ra, dec, ra_center, dec_center, pmra, pmdec):
    """Project celestial coordinates and proper motions onto the tangent plane.

    Performs a gnomonic (tangent plane) projection of sky positions centered at
    (ra_center, dec_center), and propagates proper motion vectors into the
    tangent plane coordinate system using the Jacobian of the projection.

    Parameters
    ----------
    ra : array_like
        Right ascension of the source(s) to project, in degrees.
    dec : array_like
        Declination of the source(s) to project, in degrees.
    ra_center : array_like
        Right ascension of the tangent point (projection center), in degrees.
    dec_center : array_like
        Declination of the tangent point (projection center), in degrees.
    pmra : array_like
        Proper motion in right ascension, in degrees. Should already include
        the cos(dec) factor if applicable.
    pmdec : array_like
        Proper motion in declination, in degrees.

    Returns
    -------
    tpx : ndarray
        Tangent plane x-coordinate of the projected position, in radians.
    tpy : ndarray
        Tangent plane y-coordinate of the projected position, in radians.
    pmtpx : ndarray
        Proper motion projected onto the tangent plane x-axis, in radians.
        Computed as the dot product of the gnomonic Jacobian row dx/d(ra,dec)
        with the proper motion vector (pmra, pmdec).
    pmtpy : ndarray
        Proper motion projected onto the tangent plane y-axis, in radians.
        Computed as the dot product of the gnomonic Jacobian row dy/d(ra,dec)
        with the proper motion vector (pmra, pmdec).
    """
    ra = np.deg2rad(np.asarray(ra))
    dec = np.deg2rad(np.asarray(dec))
    ra_center = np.deg2rad(np.atleast_1d(ra_center))
    dec_center = np.deg2rad(np.atleast_1d(dec_center))
    pmra = np.deg2rad(np.asarray(pmra))
    pmdec = np.deg2rad(np.asarray(pmdec))

    tpx, tpy = gnomonic_projection(ra, dec, ra_center, dec_center)
    tpdx, tpdy = gnomonic_dxy(ra, dec, ra_center, dec_center)

    tpdx = tpdx.T.reshape(-1, 1, 2)
    tpdy = tpdy.T.reshape(-1, 1, 2)
    pm = np.array([pmra, pmdec])
    pm = pm.T.reshape(-1, 2, 1)
    pmtpx = tpdx @ pm
    pmtpy = tpdy @ pm

    return tpx[0], tpy[0], pmtpx.squeeze(), pmtpy.squeeze()



import numpy as np
import pytest

from saltworks.fitparameters import FitParameters, Structure, structarray


def test_structure():
    s = Structure([('a', 7), 3])
    assert 10 == len(s)
    assert ['a', '__0__'] == [str(name) for name in s]
    assert 4 == len(Structure([('a', 3), 'b']))


def test_structarray():
    v = structarray(np.zeros(10), [('a', 3), ('b', 7)])
    assert np.allclose(np.asarray([ 0.,  0.,  0.]), v['a'])

    C = structarray(np.zeros((10,10)), [('a', 2), ('b', 8)])
    assert np.allclose(np.asarray([[0, 0], [0, 0]]), C['a', 'a'])


@pytest.fixture
def params():
    return FitParameters([('alpha', 2), ('S', 3), ('dSdT', 2), 'idark'])


def test_fitparameters_set(params):
    params['idark'][0] = -1
    params['idark'] = -2.1548
    params['S'][1] = 4
    params['dSdT'][:] = 42.

    assert len(params.full) == 8
    assert len(params.free) == 8
    assert np.allclose(
        np.asarray([0., 0., 0., 4., 0., 42., 42., -2.1548]),
        np.asarray(params.full))


def test_fitparameters_fix(params):
    params.fix(0, 30)
    params.fix(-2, 2)

    assert len(params.full) == 8
    assert len(params.free) == 6
    assert np.allclose(
        np.asarray([30., 0., 0., 0., 0., 0., 2, 0.]),
        np.asarray(params.full))


def test_fitparameters_fix_named(params):
    params['S'].fix([1, -1], 30)
    params['alpha'].fix(val=1e-3)
    params['dSdT'].fix()

    assert len(params.full) == 8
    assert len(params.free) == 2
    assert np.allclose(
        np.asarray([1e-3, 1e-3, 0., 30, 30, 0, 0, 0]),
        np.asarray(params.full))
    assert np.allclose(
        np.asarray([0., 0]),
        np.asarray(params.free))


def test_fitparameters_free(params):
    params.free[0] = 12
    assert params.free[0] == 0

    params['S'].free = 5
    assert np.allclose(
        np.asarray([0., 0., 5., 5., 5., 0., 0., 0.]),
        np.asarray(params.full))


def test_fitparameter_indexof(params):
    assert list(params['dSdT'].indexof()) == [5, 6]
    params['dSdT'].fix(1)
    assert list(params['dSdT'].indexof()) == [5, -1]



# This code originates from https://gitlab.in2p3.fr/betoule/imageproc by Marc
# Betoule.


import numpy as np
from scipy import sparse

from .gnomonic import inverse_gnomonic_dcenter as igp_dcenter
from .gnomonic import inverse_gnomonic_dradec as igp_dradec
from .gnomonic import inverse_gnomonic_projection as igp


class ParamFunc:
    def __call__(self, x, p, **keys):
        """Function call

        Should implement the evaluation of a parametric and vectorial
        function of x.

        Params:
        -------
        x: ndarray N, m
        p: FitParameters instance
        Return:
        ------
        f: ndarray M, N

        """
        pass

    def derivatives(self, x, p, **keys):
        """Function call and derivatives

        Should implement the evaluation of a parametric and vectorial
        function of x and of its derivatives wrt the variable and the
        parameters.

        Params:
        -------
        x: ndarray N, m
        p: FitParameters instance
        Return:
        ------
        f: ndarray (M,N)
        dfdx: ndarray (M,m,N) the derivatives of f wrt x
        dfdp: the stacked derivatives of f.flat wrt p, matrix of size (NxM, P) in sparse format i,j,val

        """
        pass

    def test_derivatives(self, x, p, **keys):
        Y0, dY0dx, (ii, jj, vals) = self.derivatives(x, p, **keys)
        N, M, nobj = dY0dx.shape
        dYdp = [
            sparse.coo_matrix((val, (ii, jj)), shape=(nobj, len(p.full))).tocsc()
            for val in vals
        ]
        err = []
        D = []
        Dpred = []
        for i in range(M):
            dx = 1e-8
            xdx = x.copy()
            xdx[:, i] += dx
            Yx = self(xdx, p, **keys)
            D.append(Yx - Y0)
            Dpred.append(dY0dx[:, i, :] * dx)
            err.append((Yx - Y0) - (dY0dx[:, i, :] * dx))
        for i in range(len(p.full)):
            p2 = p.copy()
            p2.full[i] += dx
            Yx = self(x, p2, **keys)
            Ypred = np.array(
                [np.array((dx * dY[:, i]).todense()).squeeze() for dY in dYdp]
            )
            D.append(Yx - Y0)
            Dpred.append(Ypred)
            err.append((Yx - Y0) - Ypred)
        return err, D, Dpred

    def stack(self, x, p, **keys):
        """Stack the result of vector function

        Can be useful if we want to use 1 dimensionnal fit routines
        """
        return self(x, p, **keys).flatten()

    def dstackdp(self, x, p, **keys):
        """Return the sparse matrix corresponding to the derivatives of the "stacked" function with respect to the named parameters."""
        X, D, (i, j, vals) = self.derivatives(x, p, **keys)
        M, N, L = D.shape
        ii = np.hstack([i + L * m for m in range(M)])
        jj = np.tile(j, M)
        val = np.hstack(vals)
        return X.flatten(), ii, jj, val


class StackedFunc(ParamFunc):
    def __init__(self, f):
        self.f = f

    def __call__(self, x, p, **keys):
        return self.f.stack(x, p, **keys)

    def get_struct(self):
        return self.f.get_struct()

    def derivatives(self, x, p, **keys):
        X, D, (i, j, vals) = self.f.derivatives(x, p, **keys)
        M, N, L = D.shape
        ii = np.hstack([i + L * m for m in range(M)])
        jj = np.tile(j, M)
        val = np.hstack(vals)
        return (
            X.flatten(),
            np.hstack([D[i, :, :] for i in range(M)])[None, ...],
            (ii, jj, [val]),
        )


class Polynomial(ParamFunc):
    def __init__(self, deg=2, coefsuffix="alpha"):
        self.deg = deg
        self.coefsuffix = coefsuffix
        self.coefnames = [coefsuffix + "x" * i for i in range(self.deg + 1)]

    def get_struct(self):
        return [(n, 1) for n in self.coefnames]

    def __call__(self, x, p):
        y = np.full(len(x), p[self.coefnames[-1]].full)
        for d in self.coefnames[-2::-1]:
            y *= x
            y += p[d].full
        return y

    def derivatives(self, x, p):
        y = np.ones(len(x))
        a = np.arange(len(x))
        d = self.coefnames[0]
        val = [y.copy()]
        i = [a]
        j = [np.full(len(x), p[d].indexof())]
        f = y * p[d].full
        df = 0
        for _d, d in enumerate(self.coefnames[1:], 1):
            i.append(a)
            j.append(np.full(len(x), p[d].indexof()))
            coef = p[d].full
            df += coef * y * _d
            y *= x
            val.append(y.copy())
            f += coef * y
        return f, df, (np.hstack(i), np.hstack(j), np.hstack(val))


class BiPol2D(ParamFunc):
    def __init__(self, deg=2, suffix=["alpha", "beta"], key="ccd", n=1):
        self.deg = deg
        self.suffix = suffix
        self.key = key
        self.coeffs = [(i, j) for j in range(0, deg + 1) for i in range(0, deg + 1 - j)]
        self.coeffnames = [
            suf + "x" * i + "y" * j for i, j in self.coeffs for suf in self.suffix
        ]
        self.n = n

    def get_struct(self):
        return [(name, self.n) for name in self.coeffnames]

    def __call__(self, xy, p, **keys):
        x, y = xy
        k = keys[self.key]
        f = [0] * len(self.suffix)
        for degx, degy in self.coeffs:
            monom = x**degx * y**degy
            for _s, suf in enumerate(self.suffix):
                name = suf + "x" * degx + "y" * degy
                coef = p[name].full[k]
                f[_s] = coef * monom + f[_s]
        return np.array(f)

    def derivatives(self, xy, p, **keys):
        x, y = xy
        k = keys[self.key]
        f = [0] * len(self.suffix)
        df = [np.zeros(xy.shape)] * len(self.suffix)
        i = np.tile(np.arange(len(x)), len(self.coeffnames))
        j = []
        vals = [list() for _s in self.suffix]
        for degx, degy in self.coeffs:
            monom = x**degx * y**degy
            for _s, suf in enumerate(self.suffix):
                name = suf + "x" * degx + "y" * degy
                j.append(p[name].indexof(k))
                coef = p[name].full[k]
                f[_s] = coef * monom + f[_s]
                for pouet, tralala in enumerate(self.suffix):
                    if pouet == _s:
                        vals[_s].append(monom)
                    else:
                        vals[pouet].append(np.zeros(len(monom)))
                df[_s] = (
                    np.array(
                        [
                            degx * coef * x ** (degx - 1) * y**degy,
                            degy * coef * x**degx * y ** (degy - 1),
                        ]
                    )
                    + df[_s]
                )
        for v in range(len(vals)):
            vals[v] = np.hstack(vals[v])
        return np.array(f), np.array(df), (i, np.hstack(j), vals)


class InverseGnomonic(ParamFunc):
    def __init__(self, param_names=["rac_rad", "decc_rad"], key="exp"):
        self.ra0, self.dec0 = param_names
        self.key = key

    def __call__(self, XY, p, **keys):
        piece = keys[self.key]
        rac = p[self.ra0].full[piece]
        decc = p[self.dec0].full[piece]
        X, Y = XY
        return np.array(igp(X, Y, rac, decc))

    def get_struct(self, n):
        return [(name, n) for name in [self.ra0, self.dec0]]

    def derivatives(self, XY, p, **keys):
        piece = keys[self.key]
        rac = p[self.ra0].full[piece]
        decc = p[self.dec0].full[piece]
        X, Y = XY
        N = len(X)
        (dradrac, draddecc), (ddecdrac, ddecddecc) = igp_dcenter(X, Y, rac, decc)
        valra = np.hstack([dradrac, draddecc])
        valdec = np.hstack([ddecdrac, ddecddecc])
        ii = np.tile(np.arange(N), 2)
        jj = np.hstack([p[self.ra0].indexof(piece), p[self.dec0].indexof(piece)])
        (dradX, dradY), (ddecdX, ddecdY) = igp_dradec(X, Y, rac, decc)
        return (
            self(XY, p, **keys),
            np.array([[dradX, dradY], [ddecdX, ddecdY]]),
            (ii, jj, (valra, valdec)),
        )


class D32(ParamFunc):
    def __init__(self, param_names=["rac_rad", "decc_rad"]):
        self.ra0, self.dec0 = param_names

    def __call__(self, XY, p, piece=0):
        X, Y = XY[0, :], XY[1, :]
        return np.array([X, 2 * Y, np.full(len(X), 3 * p["rac_rad"].full[piece])])

    def derivatives(self, XY, p, piece=0):
        N = XY.shape[1]
        ii = np.tile(np.arange(N), 1)
        jj = np.full(N, p["rac_rad"].indexof(piece))
        val = np.full(N, 3)
        return (
            self(XY, p, piece),
            np.array(
                [
                    [np.ones(N), np.zeros(N)],
                    [np.zeros(N), np.full(N, 2)],
                    [np.zeros(N), np.zeros(N)],
                ]
            ),
            (ii, jj, (np.zeros(N), np.zeros(N), val)),
        )


class D23(ParamFunc):
    def __init__(self, param_names=["rac_rad", "decc_rad"]):
        self.ra0, self.dec0 = param_names

    def __call__(self, XY, p, piece=0):
        X, Y, Z = XY
        return np.array([p["decc_rad"].full[piece] * X, Y * Z])

    def derivatives(self, XY, p, piece=0):
        N = XY.shape[1]
        X, Y, Z = XY
        ii = np.tile(np.arange(N), 1)
        jj = np.full(N, p["decc_rad"].indexof(piece))
        val = np.full(N, 3)
        return (
            self(XY, p, piece),
            np.array(
                [
                    [np.full(N, p["decc_rad"].full[piece]), np.zeros(N), np.zeros(N)],
                    [np.zeros(N), Z, Y],
                ]
            ),
            (ii, jj, (val, np.zeros(N))),
        )


class CompoundFunc(ParamFunc):
    def __init__(self, f, g):
        self.f = f
        self.g = g

    def __call__(self, x, p, **keys):
        return self.f(self.g(x, p, **keys), p, **keys)

    def derivatives(self, x, p, **keys):
        X, dgdx, (idgdp, jdgdp, dgdp) = self.g.derivatives(x, p, **keys)
        Y, dfdX, (idfdp, jdfdp, dfdp) = self.f.derivatives(X, p, **keys)
        M, N, L = dfdX.shape
        CR = np.matmul(dgdx.T, dfdX.T).T
        ii = np.hstack([idgdp, idfdp])
        jj = np.hstack([jdgdp, jdfdp])
        val = [
            np.hstack(
                [
                    np.sum([dgdp[i] * dfdX[j, i, idgdp] for i in range(N)], axis=0),
                    dfdp[j],
                ]
            )
            for j in range(M)
        ]
        return Y, CR, (ii, jj, val)



"""Handles parameters for linear and non-linear fits

This module contains utility classes to handle parameters in linear and
non-linear least square fits implemented in linearmodels and nonlinearmodels.
It provide 2 main features:

- `structarray`: a derivative of numpy class ndarray to manage large vectors
organized into named subgroups.


- `FitParameters` : a class to manage large parameter vectors. It allows to
easily fix/release specific parameters or entire subgroups, and remap the
remaining free parameters into a contiguous vector.

"""

import numpy as np


class Structure:
    """Collection of named slices

    Slices are specified by a name and a length. If omitted, the length default
    to one and the name to __0__ for the first unamed slice, __1__ for the
    second and so on.

    Examples:
    ---------
    >>> s = Structure([('a', 7), 3])
    >>> print len(s)
    10
    >>> for name in s: print name
    a
    __0__
    >>> print len(Structure([('a', 3), 'b']))
    4

    """

    def __init__(self, groups):
        if isinstance(groups, Structure):
            self.slices = groups.slices.copy()
            self._len = groups._len
        else:
            self.slices = {}
            i = 0
            _n_unnamed = 0
            for group in groups:
                if isinstance(group, str):
                    name = group
                    _len = 1
                else:
                    try:
                        name, _len = group
                    except TypeError:
                        try:
                            _len = int(group)
                            name = "__%d__" % _n_unnamed
                            _n_unnamed += 1
                        except:
                            raise TypeError(
                                "Structure specification not understood: %s"
                                % repr(group)
                            )
                self.slices[name] = slice(i, i + _len)
                i += _len
            self._len = i

    def __getitem__(self, arg):
        if isinstance(arg, str):
            return self.slices[arg]
        else:
            return arg

    def __iter__(self):
        return self.slices.__iter__()

    def __len__(self):
        return self._len


class structarray(np.ndarray):
    """Decorate numpy arrays with a collection of named slices.

    Array slices becomes accessible by their name. This is applicable to
    nd array, although the same `Structure` is shared between all
    dimensions.

    Examples:
    ---------
    >>> v = structarray(np.zeros(10), [('a', 3), ('b', 7)])
    >>> print v['a']
    [ 0.  0.  0.]

    >>> C = structarray(np.zeros((10,10)), [('a', 2), ('b', 8)])
    >>> print C['a', 'a']
    [[ 0.  0.]
     [ 0.  0.]]
    """

    def __new__(cls, array, struct=[]):
        obj = array.view(cls)
        obj.struct = Structure(struct)
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.struct = getattr(obj, "struct", [])

    def __getitem__(self, args):
        if not isinstance(args, tuple):
            args = (args,)
        newargs = tuple([self.struct[arg] for arg in args])
        return np.asarray(self)[newargs]

    def __setitem__(self, args, val):
        if not isinstance(args, tuple):
            args = (args,)
        newargs = tuple([self.struct[arg] for arg in args])
        np.asarray(self)[newargs] = val

    def __getslice__(self, start, stop):
        return self.__getitem__(slice(start, stop))

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(structarray, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (self.struct,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.struct = state[-1]  # Set the info attribute
        # Call the parent's __setstate__ with the other tuple elements.
        super(structarray, self).__setstate__(state[0:-1])


class FitParameters:
    """Manages a vector of fit parameters with the possibility to mark a subset
    of them as fixed to a given value.

    The parameters can be organized in named slices (block of contiguous
    values) accessible through indexing by their name as in `structarray`.

    >>> p_salt = FitParameters(['X0', 'X1', 'Color', 'Redshift'])
    >>> p_dice = FitParameters([('alpha', 2), ('S', 10), ('dSdT', 10), 'idark'])

    It is possible to modify the parameters in place. Using the indexing of
    slices by name simplifies somewhat the operations, as one does not need to
    care about the position of a slice within the entire parameter vector:

    >>> p_dice['idark'][0] = -1.0232E-12
    >>> p_dice['S'][4] = 25.242343E-9
    >>> p_dice['dSdT'][:] = 42.
    >>> print p_dice
    alpha: array([ 0.,  0.])
    S: array([  0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
             0.00000000e+00,   2.52423430e-08,   0.00000000e+00,
             0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
             0.00000000e+00])
    dSdT: array([ 42.,  42.,  42.,  42.,  42.,  42.,  42.,  42.,  42.,  42.])
    idark: array([ -1.02320000e-12])

    It is also possible to mark parameters as fixed to a value.
    >>> p_dice.fix(0, 12.)

    Value is optional. The above is equivalent to:
    >>> p_dice[0] = 12.
    >>> p_dice.fix(0)

    Again named slices simplifies the operations:
    >>> p_dice['S'].fix([0, -1], 12.)
    >>> p_dice['dSdT'].fix([0, -1])

    One can fix entire slices at once:
    >>> p_dice['idark'].fix()
    >>> p_salt['Redshift'].fix(val=0.23432)

    The property ``full'' give access to the vector of parameters. The
    property "free" gives access to the free parameters:
    >>> print len(p_dice.free), len(p_dice.full)
    17 23

    Note that free relies on fancy indexing. Access thus trigger a
    copy. As a consequence, the following will not actually alter the
    data:
    >>> p_dice.free[0] = 12.
    >>> print p_dice.free[0]
    0.0

    It is still possible to set slices of free parameters as a
    contiguous vector. For example:
    >>> p_dice['S'].free = 12.
    >>> print p_dice['S'].free
    [ 12.  12.  12.  12.  12.  12.  12.  12.]

    >>> p_dice[:5].free = 4.
    >>> print p_dice[:5].free
    [ 4.  4.  4.]

    In particular, the typical use case which consists in updating the
    free parameters with the results of a fit works as expected:
    >>> p = np.arange(len(p_dice.free))
    >>> p_dice.free = p

    Last the class provide a convenience function that return the index
    of a subset of parameters in the global free parameters vector, and
    -1 for fixed parameters:
    >>> print p_dice['dSdT'].indexof()
    [-1  9 10 11 12 13 14 15 16 -1]
    >>> print p_dice['dSdT'].indexof([1,2])
    [ 9 10]

    """

    def __init__(self, groups):
        struct = Structure(groups)
        self._free = structarray(np.ones(len(struct), dtype="bool"), struct)
        self._pars = structarray(np.zeros(len(struct), dtype="float"), struct)
        self._index = structarray(np.zeros(len(struct), dtype="int"), struct)
        self._struct = struct
        self._reindex()
        self._base = self

    def copy(self):
        cop = FitParameters(self._struct)
        cop._free = structarray(np.copy(self._free), cop._struct)
        cop._pars = structarray(np.copy(self._pars), cop._struct)
        cop._index = structarray(np.copy(self._index), cop._struct)
        cop._reindex()
        cop._base = cop
        return cop

    def _reindex(self):
        self._index[self._free] = np.arange(self._free.sum())
        self._index[~self._free] = -1

    def fix(self, keys=slice(None), val=None):
        self._free[keys] = False
        if val is not None:
            self._pars[keys] = val
        self._base._reindex()

    def release(self, keys=slice(None)):
        self._free[keys] = True
        self._base._reindex()

    def indexof(self, indices=slice(None)):
        return self._index[indices]

    def __getitem__(self, args):
        # Prevent conversion to scalars
        if isinstance(args, int):
            args = slice(args, args + 1)
        new = FitParameters.__new__(FitParameters)
        new._free = self._free[args]
        new._pars = self._pars[args]
        new._index = self._index[args]
        new._base = self._base
        return new

    def __setitem__(self, args, val):
        self._pars[args] = val

    @property
    def free(self):
        return self._pars[self._free]

    @free.setter
    def free(self, val):
        self._pars[self._free] = val

    @property
    def full(self):
        return self._pars

    @full.setter
    def full(self, val):
        self._pars = val

    def __repr__(self):
        if hasattr(self, "_struct"):
            s = "\n".join(
                ["%s: %s" % (key, repr(self._pars[key])) for key in self._struct]
            )
        else:
            s = repr(self._pars)
        return s



#!/usr/bin/python

import numpy as np

deg2rad = deg2rad = np.pi / 180

def euclidian(x1, y1, x2, y2):
    return np.sqrt((np.array(x1) - np.array(x2)) ** 2 + (np.array(y1) - np.array(y2)) ** 2)

def haversine(ra1, dec1, ra2, dec2):
    dlambda = np.array(ra1 - ra2)
    return np.arccos(np.sin(dec1) * np.sin(dec2) + np.cos(dec1) * np.cos(dec2) * np.cos(dlambda))


def gnomonic_projection(ra, dec, center=None):
    ra_rad, dec_rad = ra * deg2rad, dec * deg2rad
    if center is None:
        center = ra_rad.mean(), dec_rad.mean()
    else:
        center = center[0] * deg2rad, center[1] * deg2rad
    s, c = np.sin(dec_rad), np.cos(dec_rad)
    c2 = np.cos(center[0] - ra_rad)
    d = np.sin(center[1]) * s + np.cos(center[1]) * c * c2
    x = (np.cos(center[1]) * np.sin(center[0] - ra_rad)) / d
    y = (np.sin(center[1]) * c - np.cos(center[1]) * s * c2) / d
    return x, y


class NearestNeighAssoc():
    """Solve the fixed-radius nearest neighbor search on a 2D cartesian
    lattice
    """
    def __init__(self, first=[], extension=[], radius=1):
        self.belongs = {}
        self.clusters = []
        self.radius = radius

        if extension:
            xmin, xmax, ymin, ymax = extension
            self.x_bins = np.arange(xmin - 0.01 * radius, xmax + 0.01 * radius, radius)
            self.y_bins = np.arange(ymin - 0.01 * radius, ymax + 0.01 * radius, radius)
        elif first:
            firstx, firsty = first
            xmin, xmax, ymin, ymax = firstx.min(), firstx.max(), firsty.min(), firsty.max()
            self.x_bins = np.arange(xmin - 0.01 * radius, xmax + 0.01 * radius, radius)
            self.y_bins = np.arange(ymin - 0.01 * radius, ymax + 0.01 * radius, radius)
            self.clusters = list(zip(firstx, firsty))
            i = np.digitize(firstx, self.x_bins)
            j = np.digitize(firsty, self.y_bins)
            for k in range(len(i)):
                ik, jk = i[k], j[k]
                self.belongs[(ik, jk)] = self.belongs.get((ik, jk), []) + [k]


    def append(self, x, y, metric=haversine):
        if not hasattr(self, 'x_bins'):
            xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
            self.x_bins = np.arange(xmin - 0.01 * self.radius, xmax + 0.01 * self.radius, self.radius)
            self.y_bins = np.arange(ymin - 0.01 * self.radius, ymax + 0.01 * self.radius, self.radius)

        i = np.digitize(x, self.x_bins)
        j = np.digitize(y, self.y_bins)
        index = np.zeros(len(i))

        for k in range(len(i)):
            ik, jk = i[k], j[k]
            # gather the list of clusters in the neighborhood
            candidates = sum([self.belongs.get((i_n, j_n), []) for i_n in (ik - 1, ik, ik + 1) for j_n in (jk - 1, jk, jk + 1)],[])
            if candidates:
                distance = metric(x[k], y[k], [self.clusters[l][0] for l in candidates], [self.clusters[l][1] for l in candidates])
                l = distance.argmin()
            if candidates and distance[l] < self.radius:
                m = candidates[l]
                index[k] = m
                self.clusters[m][2] += 1
            else:
                clu_i = len(self.clusters)
                index[k] = clu_i
                self.clusters.append([x[k], y[k], 1])
                self.belongs[(ik, jk)] = self.belongs.get((ik, jk), []) + [clu_i]
        return index

    def match(self, x, y, metric=haversine):
        i = np.digitize(x, self.x_bins)
        j = np.digitize(y, self.y_bins)
        index = np.zeros(len(i), dtype='int')
        for k in range(len(i)):
            ik, jk = i[k], j[k]
            # gather the list of clusters in the neighborhood
            candidates = sum([self.belongs.get((i_n, j_n), []) for i_n in (ik - 1, ik, ik + 1) for j_n in (jk - 1, jk, jk + 1)],[])
            if candidates:
                distance = metric(x[k], y[k], [self.clusters[l][0] for l in candidates], [self.clusters[l][1] for l in candidates])
                l = distance.argmin()
            if candidates and distance[l] < self.radius:
                m = candidates[l]
                index[k] = m
            else:
                index[k] = -1
        return index

    def get_cat(self):
        clusters = np.rec.fromrecords(self.clusters, names=['ra', 'dec', 'n'])
        clusters['ra'] /= deg2rad
        clusters['dec'] /= deg2rad
        return clusters

def assoc(catalog, project=True, xy=False, radius=3e-4 * deg2rad):
    if project:
        x, y = gnomonic_projection(np.array(catalog['ra']), np.array(catalog['dec']))
#        x, y = gnomonic_projection(catalog['ra'], catalog['dec'])
    elif xy:
        x, y = np.array(catalog['x']), np.array(catalog['y'])
    else:
        x, y = stack['ra'] * deg2rad, stack['dec'] * deg2rad
    assoc =  NearestNeighAssoc(radius=radius)
    index = assoc.append(x, y, metric=euclidian if (project or xy) else haversine)
    clusters = assoc.get_cat()
    return index.astype(int), clusters

def match(refcat, cat, project=True, xy=False, arcsecrad=1):
    def _arr(x):
        return np.array(x)
    if project:
        xref, yref = gnomonic_projection(_arr(refcat['ra']), _arr(refcat['dec']),
                                         center=[refcat['ra'].mean(), refcat['dec'].mean()])
        x, y = gnomonic_projection(_arr(cat['ra']), _arr(cat['dec']),
                                   center=[refcat['ra'].mean(), refcat['dec'].mean()])
    elif xy:
        xref, yref = _arr(refcat["x"]), _arr(refcat["y"])
        x, y = _arr(cat["x"]), _arr(cat["y"])
    else:
        xref, yref = _arr(refcat['ra']) * deg2rad, _arr(refcat['dec']) * deg2rad
        x, y = _arr(cat['ra']) * deg2rad, _arr(cat['dec']) * deg2rad
    assoc =  NearestNeighAssoc(first=[xref, yref], radius=arcsecrad/3600. * deg2rad)
    index = assoc.match(x, y, metric=euclidian if (project or xy) else haversine)
    return index

def radius_to_nside(radius):
    """Find the smallest nside for wich a circle of radius radius is
    contained within an healpy pixel.
    
    radius: in radian
    """
    npix = 4 * np.pi / (2 * radius) ** 2
    order = int(np.ceil(np.log2(np.sqrt(npix / 12))))
    return 2 ** order

class NearestNeighAssocHealpy():
    """Solve the fixed-radius nearest neighbor search on a 2D Healpy
    lattice. Good for full-sky match.

    """

    def __init__(self, first=[], radius=np.pi / 180./ 3600.):
        import healpy
        self.belongs = {}
        self.clusters = []
        self.radius = radius
        self.nside = radius_to_nside(radius)

        if first:
            firstra, firstdec = first[0] * deg2rad, first[1] * deg2rad
            self.clusters = list(zip(firstra, firstdec))
            i = healpy.ang2pix(self.nside, np.pi/2 - firstdec, firstra)
            for k in range(len(i)):
                self.belongs[i[k]] = self.belongs.get(i[k], []) + [k]


    def append(self, ra, dec, metric=haversine):
        import healpy
        ra_rad, dec_rad = np.array(ra * deg2rad), np.array(dec * deg2rad)
        i = healpy.ang2pix(self.nside, np.pi/2 - dec_rad, ra_rad)
        index = np.zeros(len(i), dtype='int')

        for k in range(len(i)):
            ik = i[k]
            # gather the list of clusters in the neighborhood
            neighpix = list(healpy.get_all_neighbours(self.nside, ik)) + [ik]
            candidates = sum([self.belongs.get(i_n, []) for i_n in neighpix if neighpix != -1],[])
            if candidates:
                distance = metric(ra_rad[k], dec_rad[k], [self.clusters[l][0] for l in candidates], [self.clusters[l][1] for l in candidates])
                l = distance.argmin()
            if candidates and distance[l] < self.radius:
                m = candidates[l]
                index[k] = m
                self.clusters[m][2] += 1
            else:
                clu_i = len(self.clusters)
                index[k] = clu_i
                self.clusters.append([ra_rad[k], dec_rad[k], 1])
                self.belongs[ik] = self.belongs.get(ik, []) + [clu_i]
        return index

    def match(self, ra, dec, metric=haversine):
        import healpy
        ra_rad, dec_rad = np.array(ra * deg2rad), np.array(dec * deg2rad)
        i = healpy.ang2pix(self.nside, np.pi/2 - dec_rad, ra_rad)
        index = np.zeros(len(i), dtype='int')
        for k in range(len(i)):
            ik = i[k]
            # gather the list of clusters in the neighborhood
            neighpix = list(healpy.get_all_neighbours(self.nside, ik)) + [ik]
            candidates = sum([self.belongs.get(i_n, []) for i_n in neighpix if neighpix != -1],[])
            if candidates:
                distance = metric(ra_rad[k], dec_rad[k], [self.clusters[l][0] for l in candidates], [self.clusters[l][1] for l in candidates])
                l = distance.argmin()
            if candidates and distance[l] < self.radius:
                m = candidates[l]
                index[k] = m
            else:
                index[k] = -1
        return index

    def get_cat(self):
        clusters = np.rec.fromrecords(self.clusters, names=['ra', 'dec', 'n'])
        clusters['ra'] /= deg2rad
        clusters['dec'] /= deg2rad
        return clusters



"""
Tools to read SNLS catalogs, extracted from croaks.NTuple
"""

import numpy as np


def read_header(fh):
    """ Return keys, values from file header. 
    """
    keys = {}
    names = []
    for line in fh:
        if line.startswith("#"):
            if line[1:].strip() == "end":
                break
            names.append(line[1:].split(":")[0].strip())
        elif line.startswith("@"):
            l = line[1:].split()
            keys[l[0]] = convert(" ".join(l[1:]))
    return keys, names


def convert(value):
    """ Convert into python type. 
    """
    value = value.strip()
    if not value:
        value = "nan"
    try:
        value = int(value)
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            pass
    return value


def fromtxt(filename):
    """Return catalog as recarray."""
    #pylint: disable=unspecified-encoding
    comments = set(["#", "\n"])
    with open(filename, "r") as fid:
        keys, names = read_header(fid)
        records = []
        for line in fid:
            if line[0] in comments:
                continue
            vals = line.split()
            records.append([convert(v) for v in vals])
    nt = np.rec.fromrecords(records, names=names)  # .view(NTuple)
    nt.keys = keys
    return nt



"""Indexing facilities functions"""

import numpy as np


def make_index(index):
    s = np.argsort(index)
    n = np.bincount(index.astype("int"))
    n = n[n != 0]
    l = []
    p = 0
    for i in n:
        l.append(s[p : p + i])
        p = p + i
    return l


def first_two_cols(ntuple):
    """Returns the first two columns of an array or NTuple"""
    if len(ntuple.shape) == 1:
        col_x, col_y = ntuple.dtype.names[:2]
        return ntuple[col_x], ntuple[col_y]
    else:
        return ntuple[:, 0], ntuple[:, 1]


def find_closest(data, target, sorted=True):
    """Find index closest to target from a sorted list A

    If sorted is False, takes care of it

    """

    if not sorted:
        idx_sorted = np.argsort(data)
        A = data[idx_sorted]
    else:
        A = data
        idx_sorted = np.arange(len(data) - 1)

    idx = A.searchsorted(target)
    idx = np.clip(idx, 1, len(A) - 1)
    left = A[idx - 1]
    right = A[idx]
    idx -= target - left < right - target
    if idx > len(idx_sorted) - 1:
        idx = -1
    return idx_sorted[idx]



# SPDX-FileCopyrightText: 2024-present Leander (lpnlp202111) <leander.lacroix.dev@protonmail.com>
#
# SPDX-License-Identifier: MIT

from importlib.metadata import version

from .dataproxy import DataProxy
from .fitparameters import FitParameters
from .poloka_list import PolokaListFile

__all__ = ["__version__", "DataProxy", "FitParameters", "PolokaListFile"]

__version__ = version("saltworks")



# This code originates from https://github.com/lacroixle/ztfsmp/ by Leander
# Lacroix

import gzip
from collections.abc import Iterable

import numpy as np
import pandas as pd


class PolokaListFile:
    """A parser for poloka *.list files"""

    def __init__(self, header, df, df_desc=None, df_format=None, filename=None):
        if not df_desc:
            df_desc = {}

        self.header = header
        self.df = df
        self.df_desc = df_desc
        self.df_format = df_format
        self.filename = filename

    @classmethod
    def from_filename(cls, filename, delim_whitespace=True):
        with open(filename) as f:
            header, df, df_desc, df_format = cls._read_list_ext(
                f, delim_whitespace=delim_whitespace
            )

        return cls(header, df, df_desc, df_format, filename=filename)

    @classmethod
    def from_gzip(cls, filename, delim_whitespace=True):
        with gzip.open(filename, "rt") as f:
            header, df, df_desc, df_format = cls._read_list_ext(
                f, delim_whitespace=delim_whitespace
            )

        return cls(header, df, df_desc, df_format, filename=filename)

    def write_to(self, filename):
        filename = str(filename)  # from Path if needed
        with open(filename, "w") as f:
            if self.header:
                for key in self.header.keys():
                    if isinstance(self.header[key], Iterable) and not isinstance(
                        self.header[key], str
                    ):
                        f.write(
                            "@{} {}\n".format(
                                key.upper(), " ".join(map(str, self.header[key]))
                            )
                        )
                    else:
                        f.write(f"@{key.upper()} {str(self.header[key])}\n")

            for column in self.df.columns:
                if column in self.df_desc.keys():
                    f.write(f"#{column} : {self.df_desc[column]}\n")
                else:
                    f.write(f"#{column} :\n")

            if self.df_format is not None:
                f.write(f"# format {self.df_format}\n")

            f.write("# end\n")
            self.df.to_csv(f, sep=" ", index=False, header=False)

    @staticmethod
    def _parse_header_line(line):
        line = line.split()
        key = line[0][1:].lower()

        try:
            value = np.asarray(line[1:], dtype=int).tolist()
        except ValueError:  # not castable to int, try float
            try:
                value = np.asarray(line[1:], dtype=float).tolist()
            except ValueError:  # not castable as float, keep it as str
                value = line[:1]
        if len(value) == 1:
            value = value[0]

        return key, value

    @classmethod
    def _read_list_ext(cls, f, delim_whitespace=True):
        # Extract global @ parameters
        header = {}
        line = f.readline().strip()
        curline = 1
        while line[0] == "@":
            key, value = cls._parse_header_line(line)
            header[key] = value

            line = f.readline().strip()
            curline += 1

        # Extract dataframe
        # First, column names
        columns = []
        df_format = None
        df_desc = {}
        while line[0] == "#":
            curline += 1
            line = line[1:-1].strip()

            splitted = line.split(" ")
            if splitted[0].strip() == "end":
                break

            if splitted[0].strip() == "format":
                df_format = str(" ".join(line.split(" ")[1:])).strip()
                line = f.readline()
                continue

            splitted = line.split(":")
            column = splitted[0].strip()
            columns.append(column)
            if len(splitted) > 1:
                df_desc[column] = splitted[1].strip()

            line = f.readline()

        df = pd.read_csv(
            f,
            sep=r"\s+" if delim_whitespace else " ",
            names=columns,
            index_col=False,
            skipinitialspace=True,
        )

        return header, df, df_desc, df_format



import matplotlib
import matplotlib.collections as mcoll
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats

from .indextools import make_index
from .robuststat import mad, robust_average

color_cycle = [(166, 206, 227), (31, 120, 180), (178, 223, 138), (51, 160, 44)]
color_cycle = [(r / 256.0, g / 256.0, b / 256.0) for r, g, b in color_cycle]


def binplot(
    x,
    y,
    nbins=10,
    robust=False,
    data=True,
    scale=True,
    bins=None,
    weights=None,
    ls="none",
    dotkeys={"color": "k"},
    xerr=True,
    ax=None,
    **keys,
):
    """Bin the y data into n bins of x and plot the average and
    dispersion of each bins.

    Arguments:
    ----------
    nbins: int
      Number of bins

    robust: bool
      If True, use median and nmad as estimators of the bin average
      and bin dispersion.

    data: bool
      If True, add data points on the plot

    scale: bool
      Whether the error bars should present the error on the mean or
      the dispersion in the bin

    bins: list
      The bin definition

    weights: array(len(x))
      If not None, use weights in the computation of the mean.
      Provide 1/sigma**2 for optimal weighting with Gaussian noise

    dotkeys: dict
      To keys to pass to plot when drawing data points

    ax: matplotlib axes instance. If None plot to the current axes

    **keys:
      The keys to pass to plot when drawing bins

    Exemples:
    ---------
    >>> x = np.arange(1000); y = np.random.rand(1000);
    >>> binplot(x,y)
    """
    ind = ~np.isnan(x) & ~np.isnan(y)
    x = x[ind]
    y = y[ind]
    if weights is not None:
        weights = weights[ind]
    if bins is None:
        bins = np.linspace(x.min(), x.max() + abs(x.max() * 1e-7), nbins + 1)
    ind = (x < bins.max()) & (x >= bins.min())
    x = x[ind]
    y = y[ind]
    if weights is not None:
        weights = weights[ind]
    yd = np.digitize(x, bins)
    index = make_index(yd)
    ybinned = [y[e] for e in index]
    xbinned = 0.5 * (bins[:-1] + bins[1:])
    usedbins = np.array(np.sort(list(set(yd)))) - 1
    xbinned = xbinned[usedbins]
    bins = bins[usedbins + 1]
    if ax is None and "noplot" not in keys:
        ax = plt.gca()
    if data and "noplot" not in keys:
        ax.plot(x, y, ",", **dotkeys)

    if robust is True:
        yplot = [np.median(e) for e in ybinned]
        yerr = np.array([mad(e) for e in ybinned])
    elif robust:
        yres = [
            robust_average(e, sigma=None, clip=robust, mad=False, axis=0)
            for e in ybinned
        ]
        yplot = [e[0] for e in yres]
        yerr = [np.sqrt(e[3]) for e in yres]
    elif weights is not None:
        wbinned = [weights[e] for e in index]
        yplot = [np.average(e, weights=w) for e, w in zip(ybinned, wbinned)]
        if not scale:
            # yerr = np.array([np.std((e - a) * np.sqrt(w))
            #                 for e, w, a in zip(ybinned, wbinned, yplot)])
            yerr = np.array(
                [
                    np.sqrt(np.std((e - a) * np.sqrt(w)) ** 2 / sum(w))
                    for e, w, a in zip(ybinned, wbinned, yplot)
                ]
            )
        else:
            yerr = np.array(
                [np.sqrt(1 / sum(w)) for e, w, a in zip(ybinned, wbinned, yplot)]
            )
        scale = False
        print(yplot)
    else:
        yplot = [np.mean(e) for e in ybinned]
        yerr = np.array([np.std(e) for e in ybinned])

    if scale:
        yerr /= np.sqrt(np.bincount(yd)[usedbins + 1])

    if xerr:
        xerr = np.array([bins, bins]) - np.array([xbinned, xbinned])
    else:
        xerr = None
    if "noplot" not in keys:
        ax.errorbar(xbinned, yplot, yerr=yerr, xerr=xerr, ls=ls, **keys)
    return xbinned, yplot, yerr


def bplot(
    x,
    y,
    data=True,
    binkey=None,
    robust=False,
    scale=True,
    ls="none",
    marker="s",
    **keys,
):
    if x.shape != y.shape:
        raise ValueError("x and y have different length")
    if binkey is None:
        binkey = x
    ind = make_index(binkey)
    ybinned = [y[e] for e in ind]
    xbin = [x[e] for e in ind]
    if data:
        plt.plot(x, y, ",", color="k")

    if robust is True:
        yplot = [np.median(e) for e in ybinned]
        yerr = np.array([mad(e) for e in ybinned])
    elif robust:
        yres = [
            robust_average(e, sigma=None, clip=robust, mad=False, axis=0)
            for e in ybinned
        ]
        yplot = [e[0] for e in yres]
        yerr = [np.sqrt(e[3]) for e in yres]
    else:
        yplot = [np.mean(e) for e in ybinned]
        yerr = np.array([np.std(e) for e in ybinned])

    if scale:
        yerr /= np.sqrt(np.array([len(e) for e in ybinned]))
    xplot = [np.mean(e) for e in xbin]
    plt.errorbar(xplot, yplot, yerr=yerr, ls=ls, marker=marker, **keys)
    return xplot, yplot, yerr


def make_bins(x, y, nbins):
    """return binned x,y"""
    bins = np.linspace(x.min(), x.max() + abs(x.max() * 1e-7), nbins + 1)
    # ind = (x < bins.max()) & (x >= bins.min())
    # x = x[ind]
    # y = y[ind]
    yd = np.digitize(x, bins)
    index = make_index(yd)
    xbinned = 0.5 * (bins[:-1] + bins[1:])
    usedbins = np.array(np.sort(list(set(yd)))) - 1
    xbinned = xbinned[usedbins]
    bins = bins[usedbins + 1]
    xerr = np.array([bins, bins]) - np.array([xbinned, xbinned])
    return bins, xbinned, xerr, index


def plotpolyfit(x, y, yerr=None, deg=1, stat_legend=False, return_result=False, **keys):
    """
    Uses polyfit and plots the result

    If stat_legend is set, prints the plot results in legend.
    If return_result is set, returns the polyfit p and the linspace lx
    """
    ls = keys.pop("ls", "none")
    marker = keys.pop("marker", "s")
    plt.errorbar(x, y, yerr=yerr, ls=ls, marker=marker, **keys)
    if yerr is not None:
        w = 1 / yerr**2
    else:
        w = None
    p, cov = np.polyfit(x, y, w=w, deg=deg, cov=True)
    r = y - np.polyval(p, x)

    axrange = plt.axis()[:2]
    lx = np.linspace(*axrange)
    if stat_legend:
        plt.plot(
            lx,
            np.polyval(p, lx),
            "k-",
            label=r"%3.2f ($\pm$ %2.1e) x + %3.2f ($\pm$ %2.1e)\n rms: %5.3f"
            % (p[0], cov[0][0], p[1], cov[1][1], np.std(r)),
        )
        plt.legend(prop={"size": 9})
    else:
        plt.plot(lx, np.polyval(p, lx), "k-")

    if return_result:
        return p, lx


def histofit(x, **keys):
    """Fits a 1D histogram to x

    heart=[start, end] keyword defines the heart of the distribution that will
    be fit

    """
    i_ok = np.isfinite(x)
    if np.any(~i_ok):
        print(("Found %d not-finite values in x" % (len(np.where(~i_ok)[0]))))
    x = x[i_ok]

    plt.rc("legend", fancybox=True)
    heart = keys.pop("heart", False)
    h, b, p = plt.hist(x, density=True, **keys)
    if heart:
        m, sig = scipy.stats.distributions.norm.fit(x[(x > heart[0]) & (x < heart[1])])
        n = len(x[(x > heart[0]) & (x < heart[1])])
    else:
        m, sig = scipy.stats.distributions.norm.fit(x)
        n = len(x)
    plt.plot(
        b,
        scipy.stats.distributions.norm.pdf(b, m, sig),
        label="mean: %.3g\nsig: %.3g" % (m, sig),
    )
    plt.gca().legend(loc=1, ncol=1, shadow=True)
    return m, sig, sig / np.sqrt(n)


def histofit2(x, **keys):
    # TODO why changing default ???
    plt.rc("legend", fancybox=True)
    range = keys.pop("range", [x.min(), x.max()])
    h, b, p = plt.hist(x, range=range, density=True, **keys)
    index = h > 0
    # p, e = np.polyfit(0.5 * (b[1:] + b[:-1])[index],
    #                  np.log(h[index]),
    #                  h[index] ** 2, 2)
    from .robuststat import polyfit

    p, e = polyfit((b[1:])[index], np.log(h[index]), h[index], 2)
    a, b, c = p
    sigma = 1 / np.sqrt(-2 * a)
    x0 = b * sigma**2
    A = np.exp(c + x0**2 / (2 * sigma**2))
    l = np.linspace(*range)
    plt.plot(
        l,
        A * np.exp(-((l - x0) ** 2) / (2.0 * sigma**2)),
        label="mean: %.3g\nsig: %.3g" % (x0, sigma),
    )
    plt.gca().legend(loc=1, ncol=1, shadow=True)
    return x0, sigma


def aa_plot(twocols=False, height=None):
    """Change settings to obtain plots nicely suited for A&A publications"""
    linewidth = 3.53972  # inches
    textwidth = 7.2455  # inches
    width = textwidth if twocols else linewidth
    if not height:
        height = 2.7
    plt.rc("figure", figsize=(width, height))
    # plt.rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica'], 'size': 10})
    plt.rc("font", **{"family": "serif"})
    plt.rc("axes", labelsize="small")
    # plt.rc('grid', alpha='0.5')
    plt.rc("grid", linewidth="0.1")
    plt.rc("grid", linestyle="-")
    plt.rc("grid", color="0.3")
    plt.rc("lines", markersize=3)
    plt.rc("xtick", labelsize="x-small")
    plt.rc("xtick.major", pad=2)
    plt.rc("ytick", labelsize="x-small")
    plt.rc("ytick.major", pad=2)
    plt.rc("axes", linewidth=0.1)
    plt.rc("text", usetex=True)
    # plt.rc('axes', edgecolor='0.5')
    plt.rc("text.latex", preamble=r"\usepackage{txfonts}")
    # plt.rc('ps', usedistiller='xpdf')
    plt.rc("xtick.major", size=3)
    plt.rc("xtick.minor", size=1)
    plt.rc("ytick.major", size=3)
    plt.rc("ytick.minor", size=1)
    # plt.rc('ytick', color='0.5')
    # plt.rc('xtick', color='0.5')
    plt.rcParams["xtick.direction"] = "out"
    plt.rcParams["ytick.direction"] = "out"

    plt.rc("legend", fontsize="x-small")
    # plt.rc('axes', color_cycle = color_cycle)


def density_plot(x, y, bins=(50, 50), nlev=5):
    """Density contour"""
    h, x_b, y_b = np.histogram2d(x, y, bins=bins)
    x_m = 0.5 * (x_b[:-1] + x_b[1:])
    y_m = 0.5 * (y_b[:-1] + y_b[1:])
    X, Y = np.meshgrid(x_m, y_m)
    xd = np.digitize(x, x_b) - 1
    # digitize exclude the right hand of the bin
    xd[xd == len(x_m)] = len(x_m) - 1
    yd = np.digitize(y, y_b) - 1
    yd[yd == len(y_m)] = len(y_m) - 1
    levels = np.exp(np.linspace(np.log(2), np.log(h.max()), nlev + 1))[:-1]
    C = plt.contour(X.T, Y.T, h, levels=levels)
    index = h[xd, yd] < C.levels[0]
    plt.plot(x[index], y[index], ",")


def inline_legend(axis=None, **keys):
    if axis is None:
        axis = plt.gca()
    ls = LineSet(axis)
    ls.clabel(**keys)
    plt.draw_if_interactive()


class LineSet(matplotlib.contour.ContourLabeler):
    def __init__(self, ax):
        self.axes = ax
        self.lines = [l for l in ax.lines]
        self.labelTexts = []
        self.labelCValues = []
        self.levels = [l.get_label() for l in ax.lines]
        self.cvalues = list(range(len(ax.lines)))  # [l.get_label() for l in ax.lines]
        self.collections = matplotlib.cbook.silent_list("mcoll.LineCollection")
        self.alpha = None
        self._contour_zorder = 0
        for l in self.lines:
            col = mcoll.LineCollection(
                [l.get_xydata()],
                antialiaseds=[l.get_antialiased()],
                linewidths=[l.get_linewidth()],
                linestyle=[l.get_linestyle()],
                colors=[l.get_color()],
                alpha=l.get_alpha(),
                transform=l.get_transform(),
                zorder=l.get_zorder(),
            )
            l.remove()
            self.axes.add_collection(col)
            self.collections.append(col)
        # .get_xydata()
        # self.levels = kwargs.get('levels', None)

    def to_rgba(self, cv, alpha=None):
        return matplotlib.colors.colorConverter.to_rgba(self.lines[cv].get_color())


snlsfield_color = {"D1": "r", "D2": "g", "D3": "k", "D4": "b"}


snlsfield_marker = {"D1": "+", "D2": "s", "D3": "o", "D4": "x"}


band_color = {"u": "m", "g": "g", "r": "r", "i": "k", "i2": "c", "z": "b"}


band_marker = {"u": "+", "g": "s", "r": "o", "i": "p", "i2": "D", "z": "x"}


def rstyle(ax):
    """Styles an axes to appear like ggplot2

    Must be called after all plot and axis manipulation operations have been
    carried out (needs to know final tick spacing)

    """
    # set the style of the major and minor grid lines, filled blocks
    ax.grid(True, "major", color="w", linestyle="-", linewidth=1.4)
    ax.grid(True, "minor", color="0.92", linestyle="-", linewidth=0.7)
    ax.patch.set_facecolor("0.85")
    ax.set_axisbelow(True)

    # set minor tick spacing to 1/2 of the major ticks
    ax.xaxis.set_minor_locator(
        plt.MultipleLocator((plt.xticks()[0][1] - plt.xticks()[0][0]) / 2.0)
    )
    ax.yaxis.set_minor_locator(
        plt.MultipleLocator((plt.yticks()[0][1] - plt.yticks()[0][0]) / 2.0)
    )

    # remove axis border
    for child in ax.get_children():
        if isinstance(child, matplotlib.spines.Spine):
            child.set_alpha(0)

    # restyle the tick lines
    for line in ax.get_xticklines() + ax.get_yticklines():
        line.set_markersize(5)
        line.set_color("gray")
        line.set_markeredgewidth(1.4)

    # remove the minor tick lines
    for line in ax.xaxis.get_ticklines(minor=True) + ax.yaxis.get_ticklines(minor=True):
        line.set_markersize(0)

    # only show bottom left ticks, pointing out of axis
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")

    if ax.legend_ != None:
        lg = ax.legend_
        lg.get_frame().set_linewidth(0)
        lg.get_frame().set_alpha(0.5)


def rhist(ax, data, **keywords):
    """Creates a histogram with default style parameters to look like ggplot2

    Is equivalent to calling ax.hist and accepts the same keyword parameters.
    If style parameters are explicitly defined, they will not be overwritten

    """
    defaults = {"facecolor": "0.3", "edgecolor": "0.28", "linewidth": "1", "bins": 100}

    for k, v in list(defaults.items()):
        if k not in keywords:
            keywords[k] = v

    return ax.hist(data, **keywords)


def rbox(ax, data, **keywords):
    """Creates a ggplot2 style boxplot

    It is equivalent to calling ax.boxplot with the following additions:

    Keyword arguments:
    colors -- array-like collection of colours for box fills
    names -- array-like collection of box names which are passed on as tick labels

    """
    import pylab

    hasColors = "colors" in keywords
    if hasColors:
        colors = keywords["colors"]
        keywords.pop("colors")

    if "names" in keywords:
        ax.tickNames = plt.setp(ax, xticklabels=keywords["names"])
        keywords.pop("names")

    bp = ax.boxplot(data, **keywords)
    pylab.setp(bp["boxes"], color="black")
    pylab.setp(bp["whiskers"], color="black", linestyle="solid")
    pylab.setp(bp["fliers"], color="black", alpha=0.9, marker="o", markersize=3)
    pylab.setp(bp["medians"], color="black")

    numBoxes = len(data)
    for i in range(numBoxes):
        box = bp["boxes"][i]
        boxX = []
        boxY = []
        for j in range(5):
            boxX.append(box.get_xdata()[j])
            boxY.append(box.get_ydata()[j])
        boxCoords = list(zip(boxX, boxY))

        if hasColors:
            boxPolygon = Polygon(boxCoords, facecolor=colors[i % len(colors)])
        else:
            boxPolygon = Polygon(boxCoords, facecolor="0.95")

        ax.add_patch(boxPolygon)

    return bp


def histequ_colormap(image_data, lower_bound=None, upper_bound=None):
    """Creates an histogram equalization colormap of the data

    use with cmap=histequ_colormap(image_data). Taken from
    http://stackoverflow.com/questions/5858902

    """
    data = image_data[np.isfinite(image_data)]
    image_sort = np.sort(data.flatten())
    if lower_bound is None:
        lower_bound = image_sort[0]
    if upper_bound is None:
        upper_bound = image_sort[-1]

    i_ok = (image_sort < upper_bound) & (image_sort > lower_bound)
    image_sort_ok = image_sort[i_ok]
    # See https://en.wikipedia.org/wiki/Histogram_equalization
    l_step = (image_sort_ok[:: len(image_sort_ok) / 256] - lower_bound) / (
        upper_bound - lower_bound
    )
    n_step = float(len(l_step))

    # For edge effects, the data mapping must start at 0 and go to 1
    l_step[0] = min(0.0, l_step[0])
    l_step[-1] = min(1.0, l_step[-1])

    interps = [(s, idx / n_step, idx / n_step) for idx, s in enumerate(l_step)]
    interps.append((1, 1, 1))
    cdict = {"red": interps, "green": interps, "blue": interps}
    histequ_cmap = matplotlib.colors.LinearSegmentedColormap("HistEq", cdict)

    return histequ_cmap


def wavelength_to_rgb(wavelength, gamma=0.8):
    """Converts a given wavelength of light to an approximate RGB color value

    The wavelength must be given in nanometers in the range from 380 nm through
    750 nm (789 THz through 400 THz).

    Taken from http://www.noah.org/wiki/Wavelength_to_RGB_in_Python

    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html

    Additionally alpha value set to 0.5 outside range

    """
    wavelength = float(wavelength)
    if wavelength >= 380 and wavelength <= 750:
        A = 1.0
    else:
        A = 0.5
    if wavelength < 380:
        wavelength = 380.0
    if wavelength > 750:
        wavelength = 750.0
    if wavelength >= 380 and wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        R = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        G = 0.0
        B = (1.0 * attenuation) ** gamma
    elif wavelength >= 440 and wavelength <= 490:
        R = 0.0
        G = ((wavelength - 440) / (490 - 440)) ** gamma
        B = 1.0
    elif wavelength >= 490 and wavelength <= 510:
        R = 0.0
        G = 1.0
        B = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif wavelength >= 510 and wavelength <= 580:
        R = ((wavelength - 510) / (580 - 510)) ** gamma
        G = 1.0
        B = 0.0
    elif wavelength >= 580 and wavelength <= 645:
        R = 1.0
        G = (-(wavelength - 645) / (645 - 580)) ** gamma
        B = 0.0
    elif wavelength >= 645 and wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        R = (1.0 * attenuation) ** gamma
        G = 0.0
        B = 0.0
    else:
        R = 0.0
        G = 0.0
        B = 0.0
    return (R, G, B, A)


# if __name__ == "__main__":
#     x = np.arange(1000)
#     y = np.random.randn(1000)
#     binplot(x, y)
#     plt.show()
